import logging

from p4p.server import Server

from p4pillon.definitions import PVTypes
from p4pillon.thread.pvrecipe import PVScalarRecipe

logging.basicConfig(level=logging.DEBUG)

# create an example PV of each type
# double array type PV
pvrecipe_double1 = PVScalarRecipe(PVTypes.DOUBLE, "An example double PV", 5.0)
pvrecipe_double1.initial_value = 17.5
pvrecipe_double1.description = "A different default value for the PV"
# try setting a different value for the timestamp
pvrecipe_double1.set_timestamp(1729699237.8525229)
pvrecipe_double1.set_alarm_limits(low_warning=2, high_alarm=9)

pv_double1 = pvrecipe_double1.create_pv()

Server.forever(
    providers=[
        {
            "demo:pv:name": pv_double1,  # PV name only appears here
        }
    ]
)



"""Handler for NTScalar (so far)"""

from __future__ import annotations

import logging
from collections import OrderedDict
from collections.abc import Callable

from p4p import Value
from p4p.server import ServerOperation

from p4pillon.composite_handler import AbortHandlerException
from p4pillon.server.raw import Handler, SharedPV
from p4pillon.utils import overwrite_unmarked

from .rules import (
    AlarmRule,
    BaseRule,
    ControlRule,
    ReadOnlyRule,
    RulesFlow,
    ScalarToArrayWrapperRule,
    TimestampRule,
    ValueAlarmRule,
)

logger = logging.getLogger(__name__)


class BaseRulesHandler(Handler):
    """
    Base class for handlers used to implement Normative Type handling.
    """

    def __init__(self) -> None:
        super().__init__()

        # Name is used purely for logging. Because the name of the PV is stored by
        # the Server and not the PV object associated with this handler we can't
        # determine the name until the first put operation
        self._name: str | None = None  # Used purely for logging
        self.rules: OrderedDict[str, BaseRule] = OrderedDict({"timestamp": TimestampRule()})

    def __getitem__(self, rule_name: str) -> BaseRule | None:
        """Allow access to the rules so that parameters such as read_only may be set"""
        return self.rules.get(rule_name)

    def open(self, value: Value) -> None:
        """Handler call by an open operation."""
        self._apply_rules(lambda x: x.init_rule(value))

    def post(self, pv: SharedPV, value: Value) -> None:
        """Handler call by a post operation, requires support from SharedPV derived class"""
        logger.debug("In handler post()")

        current_state = pv.current().raw  # We only need the Value from the PV
        overwrite_unmarked(current_state, value)

        self._apply_rules(lambda x: x.post_rule(current_state, value))

    def put(self, pv: SharedPV, op: ServerOperation) -> None:
        """
        Handler triggered by put operations. Note that this has additional information
        about the source of the put such as the IP address of the caller.
        """
        logger.debug("In handler put()")

        overwrite_unmarked(pv.current().raw, op.value().raw)

        rules_flow = self._apply_rules(lambda x: x.put_rule(pv, op))
        if rules_flow != RulesFlow.ABORT:
            pv.post(value=op.value())
            op.done()
        else:
            op.done(error=rules_flow.error)

    def _apply_rules(self, apply_rule: Callable[[BaseRule], RulesFlow]) -> RulesFlow:
        """
        Apply the rules. Primarily this does the basic handling of the RulesFlow.
        """

        for rule_name, rule in self.rules.items():
            logger.debug("Applying rule %s", rule_name)

            rule_flow = apply_rule(rule)

            # Originally a more elegant match (rule_flow): but we need
            # to support versions of Python prior to 3.10
            if rule_flow == RulesFlow.CONTINUE:
                pass
            elif rule_flow == RulesFlow.ABORT:
                logger.debug("Rule %s triggered handler abort", rule_name)
                return rule_flow
            elif rule_flow == RulesFlow.TERMINATE:
                logger.debug("Rule %s triggered handler terminate", rule_name)
                if "timestamp" in self.rules:
                    rule_flow = apply_rule(self.rules["timestamp"])
                return rule_flow
            elif rule_flow == RulesFlow.TERMINATE_WO_TIMESTAMP:
                logger.debug("Rule %s triggered handler terminate without timestamp", rule_name)
                return rule_flow
            else:
                logger.error("Rule %s returned unhandled return type", rule_name)
                raise TypeError(f"Rule {rule_name} returned unhandled return type {type(rule_flow)}")

        return RulesFlow.CONTINUE

    def set_read_only(self, read_only: bool = True, read_only_rule: BaseRule = ReadOnlyRule()):
        """
        Make this PV read only.
        If read_only == False then the PV is made writable
        A rule to replace the default ReadOnlyRule implementation may be passed in
        """
        if read_only:
            # Switch on the read-only rule and make sure it's the first rule
            self.rules["read_only"] = read_only_rule
            self.rules.move_to_end("read_only", last=False)
        else:
            # Switch off the read-only rule by deleting it
            self.rules.pop("read_only", None)


class ComposeableRulesHandler(Handler):
    """
    Convert the Rules interface to a simple Handler interface.
    """

    def __init__(self, rule: BaseRule) -> None:
        super().__init__()
        self.rule = rule

    def open(self, value: Value) -> None:
        """Handler call by an open operation."""
        logger.debug("In handler open()")
        self.rule.init_rule(value)

    def post(self, pv: SharedPV, value: Value) -> None:
        """Handler call by a post operation, requires support from SharedPV derived class"""
        logger.debug("In handler post()")

        current_state = pv.current().raw  # We only need the Value from the PV
        overwrite_unmarked(current_state, value)

        self.rule.post_rule(current_state, value)

    def put(self, pv: SharedPV, op: ServerOperation) -> None:
        """
        Handler triggered by put operations. Note that this has additional information
        about the source of the put such as the IP address of the caller.
        """
        logger.debug("In handler put()")

        overwrite_unmarked(pv.current().raw, op.value().raw)

        rules_flow = self.rule.put_rule(pv, op)
        if rules_flow == RulesFlow.ABORT:
            raise AbortHandlerException(rules_flow.error)

    @property
    def read_only(self) -> bool:
        """
        Set rule as resd_only.
        """
        return self.rule.read_only

    @read_only.setter
    def read_only(self, read_only: bool):
        self.rule.read_only = read_only


class NTScalarRulesHandler(BaseRulesHandler):
    """
    Rules handler for NTScalar PVs.
    """

    def __init__(self) -> None:
        super().__init__()

        self.rules["control"] = ControlRule()
        self.rules["alarm"] = AlarmRule()
        self.rules["alarm_limit"] = ValueAlarmRule()
        self.rules.move_to_end("timestamp")


class NTScalarArrayRulesHandler(BaseRulesHandler):
    """
    Rules handler for NTScalarArray PVs.
    """

    def __init__(self) -> None:
        super().__init__()

        self.rules["control"] = ScalarToArrayWrapperRule(ControlRule())
        self.rules["alarm"] = AlarmRule()  # ScalarToArrayWrapperRule unnecessary - no access to values
        self.rules["alarm_limit"] = ScalarToArrayWrapperRule(ValueAlarmRule())
        self.rules.move_to_end("timestamp")


class NTEnumRulesHandler(BaseRulesHandler):
    """
    Rules handler for NTScalarArray PVs.
    """



"""
Test p4pillon/handler.py
WARNING: AI generated code, replace with proper test cases.
"""

import unittest
from collections import OrderedDict
from unittest.mock import MagicMock

from p4pillon.composite_handler import AbortHandlerException, CompositeHandler


class DummyHandler:
    def __init__(self):
        self.calls = []

    def open(self, value):
        self.calls.append(("open", value))

    def put(self, pv, op):
        self.calls.append(("put", pv, op))

    def post(self, pv, value):
        self.calls.append(("post", pv, value))

    def rpc(self, pv, op):
        self.calls.append(("rpc", pv, op))

    def onFirstConnect(self, pv):
        self.calls.append(("onFirstConnect", pv))

    def close(self, pv):
        self.calls.append(("close", pv))


class DummyPV:
    def post(self, value):
        pass


class DummyOp:
    def __init__(self):
        self.done = MagicMock()

    def value(self):
        return None


class DummyValue:
    pass


class TestCompositeHandler(unittest.TestCase):
    def setUp(self):
        self.h1 = DummyHandler()
        self.h2 = DummyHandler()
        self.handlers = OrderedDict([("h1", self.h1), ("h2", self.h2)])
        self.comp = CompositeHandler(self.handlers)
        self.pv = DummyPV()
        self.op = DummyOp()
        self.value = DummyValue()

    def test_init_no_handlers(self):
        comp = CompositeHandler()
        with self.assertRaises(KeyError):
            comp["any"]

    def test_getitem_valid(self):
        self.assertIs(self.comp["h1"], self.h1)
        self.assertIs(self.comp["h2"], self.h2)

    def test_getitem_invalid(self):
        with self.assertRaises(KeyError):
            _ = self.comp["missing"]

    def test_open_calls_all(self):
        self.comp.open(self.value)
        self.assertEqual(self.h1.calls[0], ("open", self.value))
        self.assertEqual(self.h2.calls[0], ("open", self.value))

    def test_open_no_handlers(self):
        comp = CompositeHandler()
        comp.open(self.value)  # Should not raise

    def test_put_calls_all(self):
        self.comp.put(self.pv, self.op)
        self.assertEqual(self.h1.calls[0][0], "put")
        self.assertEqual(self.h2.calls[0][0], "put")
        self.op.done.assert_called_once_with()

    def test_put_abort_exception(self):
        def abort_put(pv, op):
            raise AbortHandlerException("abort!")

        self.h1.put = abort_put
        self.comp.put(self.pv, self.op)
        self.op.done.assert_called_once_with(error="abort!")
        # h2 should not be called
        self.assertEqual(len(self.h2.calls), 0)

    def test_post_calls_all(self):
        self.comp.post(self.pv, self.value)
        self.assertEqual(self.h1.calls[0], ("post", self.pv, self.value))
        self.assertEqual(self.h2.calls[0], ("post", self.pv, self.value))

    def test_post_no_handlers(self):
        comp = CompositeHandler()
        comp.post(self.pv, self.value)  # Should not raise

    def test_rpc_calls_all(self):
        self.comp.rpc(self.pv, self.op)
        self.assertEqual(self.h1.calls[0][0], "rpc")
        self.assertEqual(self.h2.calls[0][0], "rpc")
        self.op.done.assert_called_once_with(error=None)

    def test_rpc_abort_exception(self):
        def abort_rpc(pv, op):
            raise AbortHandlerException("rpc abort!")

        self.h2.rpc = abort_rpc
        self.comp.rpc(self.pv, self.op)
        self.op.done.assert_called_once_with(error="rpc abort!")
        # h2 should be called, but not after abort
        self.assertEqual(self.h2.calls, [])

    def test_on_first_connect_calls_all(self):
        self.comp.on_first_connect(self.pv)
        self.assertEqual(self.h1.calls[0], ("onFirstConnect", self.pv))
        self.assertEqual(self.h2.calls[0], ("onFirstConnect", self.pv))

    def test_on_first_connect_no_handlers(self):
        comp = CompositeHandler()
        comp.on_first_connect(self.pv)  # Should not raise

    def test_onFirstConnect_deprecated(self):
        self.comp.onFirstConnect(self.pv)
        self.assertEqual(self.h1.calls[0], ("onFirstConnect", self.pv))

    def test_on_last_connect_calls_all(self):
        self.comp.on_last_connect(self.pv)
        self.assertEqual(self.h1.calls[0], ("onFirstConnect", self.pv))
        self.assertEqual(self.h2.calls[0], ("onFirstConnect", self.pv))

    def test_on_last_connect_no_handlers(self):
        comp = CompositeHandler()
        comp.on_last_connect(self.pv)  # Should not raise

    def test_onLastDisconnect_deprecated(self):
        self.comp.onLastDisconnect(self.pv)
        self.assertEqual(self.h1.calls[0], ("onFirstConnect", self.pv))

    def test_close_calls_all(self):
        self.comp.close(self.pv)
        self.assertEqual(self.h1.calls[0], ("close", self.pv))
        self.assertEqual(self.h2.calls[0], ("close", self.pv))

    def test_close_no_handlers(self):
        comp = CompositeHandler()
        comp.close(self.pv)  # Should not raise


if __name__ == "__main__":
    unittest.main()



"""
Classes to define rules for handler put and PV post operations.
The RulesFlow and BaseRule classes are interfaces. The classes below those
are implementations of the logic of Normative Type
"""

# TODO: Consider adding Authentication class / callback for puts
from __future__ import annotations

import itertools
import logging
import operator
import time
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import IntEnum, auto
from functools import wraps
from typing import Any  # Hack to type hint number types
from typing import SupportsFloat as Numeric

from p4p import Type, Value
from p4p.server import ServerOperation
from p4p.server.raw import ServOpWrap, SharedPV

from p4pillon.definitions import AlarmSeverity
from p4pillon.utils import overwrite_marked, time_in_seconds_and_nanoseconds

logger = logging.getLogger(__name__)


class RulesFlow(IntEnum):
    """
    Used by the BaseRulesHandler to control whether to continue or stop
    evaluation of rules in the defined sequence. It may also be used to
    set an error message if rule evaluation is aborted.
    """

    CONTINUE = auto()  #: Continue rules processing
    TERMINATE = auto()  #: Do not process more rules but apply timestamp and complete
    TERMINATE_WO_TIMESTAMP = auto()  #: Do not process further rules; do not apply timestamp rule
    ABORT = auto()  #: Stop rules processing and abort put

    def __init__(self, _) -> None:
        # We include an error string so that we can indicate why an ABORT
        # has been triggered
        self.error: str = ""

    def set_errormsg(self, errormsg: str) -> RulesFlow:
        """
        Set an error message to explain an ABORT.
        This function returns the class instance so it may be used in lambdas
        """
        self.error = errormsg

        return self


def check_applicable_init(func):
    """
    Decorator for `BaseRule::init_rule`. Checks `is_applicable()`
    and returns RulesFlow.CONTINUE if not True
    """

    @wraps(func)
    def wrapped_function(self: BaseRule, *args, **kwargs):
        if not self.is_applicable(args[0]):
            logger.debug("Rule %s.%s is not applicable", self._name, func.__name__)  # pylint: disable=protected-access
            return RulesFlow.CONTINUE

        return func(self, *args, **kwargs)

    return wrapped_function


def check_applicable_post(func):
    """
    Decorator for `BaseRule::post_rule`. Checks `is_applicable()`
    and returns RulesFlow.CONTINUE if not True
    """

    @wraps(func)
    def wrapped_function(self: BaseRule, currentstate: Value, newpvstate: Value):
        if not self.is_applicable(newpvstate):
            logger.debug("Rule %s.%s is not applicable", self._name, func.__name__)  # pylint: disable=protected-access
            return RulesFlow.CONTINUE

        return func(self, currentstate, newpvstate)

    return wrapped_function


def check_applicable_put(func):
    """
    Decorator for `BaseRule::put_rule`. Checks `is_applicable()`
    and returns RulesFlow.CONTINUE if not True
    """

    @wraps(func)
    def wrapped_function(self: BaseRule, *args, **kwargs):
        if not self.is_applicable(args[1].value().raw):
            logger.debug("Rule %s.%s is not applicable", self._name, func.__name__)  # pylint: disable=protected-access
            return RulesFlow.CONTINUE

        return func(self, *args, **kwargs)

    return wrapped_function


def check_applicable(func):
    """
    Decorator for `BaseRule::*_rule`. Checks `is_applicable()`
    and returns RulesFlow.CONTINUE if not True
    """

    @wraps(func)
    def wrapped_function(self: BaseRule, *args, **kwargs):
        # Determine whether we're being applied to either:
        # - init_rule (1 argument)
        # - post_rule (2 arguments, second argument is a Value)
        # - put_rule (2 arguments, second argument is a ServerOperation)
        if len(args) == 1:
            newpvstate = args[0]
        elif len(args) == 2:
            if isinstance(args[1], Value):
                newpvstate = args[1]
            elif isinstance(args[1], ServOpWrap):
                newpvstate = args[1].value().raw
            else:
                raise TypeError("Type of second argument must be either Value or ServerOperation, is", type(args[1]))

        else:
            raise TypeError(f"Expected 1 or 2 arguments, received {len(args)}")

        # Then check if applicable and if not return a CONTINUE to short-circuit this rule
        if not self.is_applicable(newpvstate):
            logger.debug("Rule %s.%s is not applicable", self._name, func.__name__)  # pylint: disable=protected-access
            return RulesFlow.CONTINUE

        # Actually wrap the function we're decorating!
        return func(self, *args, **kwargs)

    return wrapped_function


class BaseRule(ABC):
    """
    Rules to apply to a PV.
    Most rules only require evaluation against the new PV state, e.g. whether to apply a control
    limit, update a timestamp, trigger an alarm etc. This may be done by the `init_rule()`.
    Other rules need to compare against the previous state of the PV, e.g. slew limits,
    control.minStep, etc. This may be done by the `post_rule()`. And some rules need to know
    who is making the request (for authorisation purposes). The may be done by the `put_rule()`
    """

    # Two members must be implemented by derived classes:
    # - _name is a human-readable name for the rule used in error and debug messages
    # - _fields is a list of the fields within the PV structure that this rule manages
    #           and at this time is used mainly by readonly rules

    @property
    @abstractmethod
    def _name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _fields(self) -> list[str]:
        raise NotImplementedError

    # Often we want to make the fields associated with a rule readonly for put
    # operations, e.g. a put operation should not be able to change the limits
    # of a valueAlarm rule. The combination of listing fields controlled by the
    # rule and having a readonly flag allows this to be automatically handled by
    # this base class's put_rule()
    read_only: bool = False

    # TODO: Consider using lru_cache but be aware of https://rednafi.com/python/lru_cache_on_methods/
    def is_applicable(self, newpvstate: Value) -> bool:
        """Test whether the Rule should be applied."""

        # _fields is None indicates the rule always applies
        if self._fields is None:
            return True

        # Next check that all the fields required are present
        if not set(self._fields).issubset(newpvstate.keys()):
            return False

        # Then check if any of the fields required are changed
        # If they aren't changed then the rule shouldn't have anything to do!
        test_fields = deepcopy(self._fields)
        if "value" not in test_fields:
            test_fields.append("value")

        if not any(newpvstate.changed(x) for x in test_fields):
            return False

        return True

    @check_applicable_init
    def init_rule(self, newpvstate: Value) -> RulesFlow:  # pylint: disable=unused-argument
        """
        Rule that only needs to consider the potential future state of a PV.
        Consider implementing if this rule could apply to a newly initialised PV.
        """
        logger.debug("Evaluating %s.init_rule", self._name)

        return RulesFlow.CONTINUE

    @check_applicable_post
    def post_rule(self, oldpvstate: Value, newpvstate: Value) -> RulesFlow:  # pylint: disable=unused-argument
        """
        Rule that needs to consider the current and potential future state of a PV.
        Usually this will involve a post where the oldpvstate is actually the current
        state of the PV, and the newpvstate represents the changes that we would
        like to apply. This rule is often also triggered in a similar manner by a
        put in which case the newpvstate derives from the ServerOperation.
        """
        logger.debug("Evaluating %s.post_rule", self._name)

        return self.init_rule(newpvstate)

    @check_applicable_put
    def put_rule(self, pv: SharedPV, op: ServerOperation) -> RulesFlow:
        """
        Rule with access to ServerOperation information, i.e. triggered by a
        handler put. These may perform authentication / authorisation style
        operations
        """

        # oldpvstate: Value = pv.current().raw
        newpvstate: Value = op.value().raw

        logger.debug("Evaluating %s.put_rule", self._name)

        if self.read_only:
            # Mark all fields of the newpvstate (i.e. op) as unchanged.
            # This will effectively make the field read-only while allowing
            # subsequent rules to trigger and work as usual
            for field in self._fields:
                # We need to rollback the changes by making the fields that shouldn't
                # be changed equal their oldstate and marking them as unchanged.
                # The first step stops issues with evaluating rules against the newstate.
                # The second step prevents changes being made.
                for changed_field in newpvstate.changedSet():
                    if changed_field.startswith(field):
                        newpvstate[changed_field] = pv.current().raw[changed_field]
                        newpvstate.mark(changed_field, False)

        return RulesFlow.CONTINUE
        # return self.post_rule(oldpvstate, newpvstate)


class BaseScalarRule(BaseRule, ABC):
    """
    Rule to be applied to NTScalarArrays
    """


class BaseGatherableRule(BaseScalarRule, ABC):
    """
    A rule usually applicable to NTScalars must be made Gatherable if when run sequentially on an
    array the correct output of a Rule must be determined by both the current Value and the
    previous value
    """

    def gather_init(self, gathered_value: Value) -> None:
        """A gather may be optionally initialised."""

    @abstractmethod
    def gather(self, scalar_value: Value, gathered_value: Value) -> None:
        """
        Gather information from multiple individual applications of a Rule
        across the elements of a NTScalarArray.
        """


class BaseArrayRule(BaseRule, ABC):
    """
    Rule to be applied to NTScalarArrays
    """


class ReadOnlyRule(BaseScalarRule):
    """A rule which rejects all attempts to put values"""

    @property
    def _name(self) -> str:
        return "read_only"

    @property
    def _fields(self) -> list[str]:
        return []

    def put_rule(self, pv: SharedPV, op: ServerOperation) -> RulesFlow:
        return RulesFlow(RulesFlow.ABORT).set_errormsg("read-only")


class AlarmRule(BaseRule):
    """
    This class exists only to allow the alarm field, i.e. severity and
    message to be made read-only for put operations
    """

    @property
    def _name(self) -> str:
        return "alarm"

    @property
    def _fields(self) -> list[str]:
        return ["alarm"]


class TimestampRule(BaseRule):
    """Set current timestamp unless provided with an alternative value"""

    @property
    def _name(self) -> str:
        return "timestamp"

    @property
    def _fields(self) -> list[str]:
        return ["timeStamp"]

    def is_applicable(self, newpvstate: Value) -> bool:
        """
        Override the base class's rule because timeStamp changes are triggered
        by changes to any field and not just to the timeStamp field
        """
        # If nothing at all has changed then don't update the timeStamp
        # TODO: Check if this is expected behaviour for Normative Types
        if not newpvstate.changedSet():
            return False

        # Check if there is a timeStamp field to update!
        if "timeStamp" not in newpvstate.keys():
            return False

        return True

    @check_applicable_init
    def init_rule(self, newpvstate: Value) -> RulesFlow:
        """Update the timeStamp of a PV"""

        seconds, nanoseconds = time_in_seconds_and_nanoseconds(time.time())
        # TODO: there's a bug in the _wrap which means that timestamps are always marked as changed
        #       Fix this when that bug is fixed.
        # if "timeStamp.secondsPastEpoch" not in newpvstate.changedSet():
        if True:
            logger.debug("using secondsPastEpoch from time.time()")
            newpvstate["timeStamp.secondsPastEpoch"] = seconds
        # if "timeStamp.nanoseconds" not in newpvstate.changedSet():
        if True:
            newpvstate["timeStamp.nanoseconds"] = nanoseconds
            logger.debug("using nanoseconds from time.time()")

        return RulesFlow.CONTINUE


class ControlRule(BaseScalarRule):
    """
    Apply rules implied by Normative Type control field.
    These include a minimum value change (control.minStep) and upper
    and lower limits for values (control.limitHigh and control.limitLow)
    """

    @property
    def _name(self) -> str:
        return "control"

    @property
    def _fields(self) -> list[str]:
        return ["control"]

    @check_applicable_init
    def init_rule(self, newpvstate: Value) -> RulesFlow:
        """Check whether a value should be clipped by the control limits

        NOTE: newpvstate from a put is a combination of the old and new state

        Returns None if no change should be made and the value is valid

        TODO: see if this can be separated out into a function like the
        min_step_violated to work better with arrays

        """
        logger.debug("Evaluating control.init rule")

        # Check lower and upper control limits
        if newpvstate["value"] < newpvstate["control.limitLow"]:
            newpvstate["value"] = newpvstate["control.limitLow"]
            logger.debug(
                "Lower control limit exceeded, changing value to %s",
                newpvstate["value"],
            )
            return RulesFlow.CONTINUE

        if newpvstate["value"] > newpvstate["control.limitHigh"]:
            newpvstate["value"] = newpvstate["control.limitHigh"]
            logger.debug(
                "Upper control limit exceeded, changing value to %s",
                newpvstate["value"],
            )
            return RulesFlow.CONTINUE

        return RulesFlow.CONTINUE

    @check_applicable_post
    def post_rule(self, oldpvstate: Value, newpvstate: Value) -> RulesFlow:
        logger.debug("Evaluating control.post rule")
        # Check minimum step first - if the check for the minimum step fails then we continue
        # and ignore the actual evaluation of the limits
        if __class__.min_step_violated(
            newpvstate["value"],
            oldpvstate["value"],
            newpvstate["control.minStep"],
        ):
            logger.debug("<minStep")
            newpvstate["value"] = oldpvstate["value"]

        # if the min step isn't violated, we continue and evaluate the limits themselves
        # on the value
        return self.init_rule(newpvstate)

    @classmethod
    def min_step_violated(cls, new_val, old_val, min_step) -> Numeric:
        """Check whether the new value is too small to pass a minStep threshold"""
        if old_val is None or min_step is None:
            return False

        return abs(new_val - old_val) < min_step


class ValueAlarmRule(BaseGatherableRule):
    """
    Rule to check whether valueAlarm limits have been triggered, changing
    alarm.severity and alarm.message appropriately.

    TODO: Implement hysteresis
    """

    @property
    def _name(self) -> str:
        return "valueAlarm"

    @property
    def _fields(self) -> list[str]:
        return ["alarm", "valueAlarm"]

    @check_applicable_init
    def init_rule(self, newpvstate: Value) -> RulesFlow:
        """Evaluate alarm value limits"""
        # TODO: Apply the rule for hysteresis. Unfortunately I don't understand the
        # explanation in the Normative Types specification...
        logger.debug("Evaluating %s.init_rule", self._name)

        # Check if valueAlarms are present and active!
        if not newpvstate["valueAlarm.active"]:
            # TODO: This is wrong! If valueAlarm was active and then made inactive
            #       the alarm will not be cleared
            logger.debug("\tvalueAlarm not active")
            return RulesFlow.CONTINUE

        try:
            # The order of these tests is defined in the Normative Types document
            if self.__alarm_state_check(newpvstate, "highAlarm"):
                return RulesFlow.CONTINUE
            if self.__alarm_state_check(newpvstate, "lowAlarm"):
                return RulesFlow.CONTINUE
            if self.__alarm_state_check(newpvstate, "highWarning"):
                return RulesFlow.CONTINUE
            if self.__alarm_state_check(newpvstate, "lowWarning"):
                return RulesFlow.CONTINUE
        except SyntaxError:
            # TODO: Need more specific error than SyntaxError and to decide
            # if continue is the correct behaviour
            return RulesFlow.CONTINUE

        # If we made it here then there are no alarms or warnings and we need to indicate that
        # possibly by resetting any existing ones
        alarms_changed = False
        if newpvstate["alarm.severity"]:
            newpvstate["alarm.severity"] = 0
            alarms_changed = True
        if newpvstate["alarm.message"]:
            newpvstate["alarm.message"] = ""
            alarms_changed = True

        if alarms_changed:
            logger.debug(
                "Setting to severity %i with message '%s'",
                newpvstate["alarm.severity"],
                newpvstate["alarm.message"],
            )
        else:
            logger.debug("Made no automatic changes to alarm state.")

        return RulesFlow.CONTINUE

    @classmethod
    def __alarm_state_check(cls, pvstate: Value, alarm_type: str, op=None) -> bool:
        """Check whether the PV should be in an alarm state"""
        if not op:
            if alarm_type.startswith("low"):
                op = operator.le
            elif alarm_type.startswith("high"):
                op = operator.ge
            else:
                raise SyntaxError(f"CheckAlarms/alarmStateCheck: do not know how to handle {alarm_type}")

        severity = pvstate[f"valueAlarm.{alarm_type}Severity"]
        if op(pvstate["value"], pvstate[f"valueAlarm.{alarm_type}Limit"]) and severity:
            pvstate["alarm.severity"] = severity

            # TODO: Understand this commented out code!
            #       I think it's to handle the case of an INVALID alarm or manually
            #       set alarm message?
            # if not pvstate.changed("alarm.message"):
            #     pvstate["alarm.message"] = alarm_type
            pvstate["alarm.message"] = alarm_type

            logger.debug(
                "Setting to severity %i with message '%s'",
                severity,
                pvstate["alarm.message"],
            )

            return True

        return False

    def gather_init(self, gathered_value: Value) -> None:
        if (
            not gathered_value.changed("alarm.severity")
            or gathered_value["alarm.severity"] != AlarmSeverity.INVALID_ALARM
        ):
            gathered_value["alarm.severity"] = AlarmSeverity.NO_ALARM
            gathered_value["alarm.message"] = ""

    def gather(self, scalar_value: Value, gathered_value: Value) -> None:
        if scalar_value["alarm.severity"] > gathered_value["alarm.severity"]:
            gathered_value["alarm.severity"] = scalar_value["alarm.severity"]
            gathered_value["alarm.message"] = scalar_value["alarm.message"]


class ScalarToArrayWrapperRule(BaseArrayRule):
    """
    Wrap a rule designed to be applied to an NTScalar so that it works with
    NTScalarArrays.
    """

    @property
    def _name(self) -> str:
        return self._wrap_name

    @property
    def _fields(self) -> list[str]:
        return self._wrap_fields

    def __init__(self, to_wrap: BaseScalarRule | BaseGatherableRule) -> None:
        super().__init__()

        self._wrapped = to_wrap

        self._wrap_name = to_wrap._name
        self._wrap_fields = to_wrap._fields

    def _get_value_id(self, arrayval: Value) -> str:
        return arrayval.type().aspy()[1]  # id of the structure, probably "epics:nt/NTScalarArray:1.0"

    def _change_array_type_to_scalar_type(self, arrayval: Value) -> list:
        """
        Return the id and type of an NTScalarArray Value, changing the type of the
        value field to be a scalar.
        """

        # The type of the scalar is essentially the same as the array with
        # the value type modified. Extracting the type info of the input value
        # and then making a change to it is surprisingly complicated!
        val_aspy = arrayval.type().aspy()
        val_type = dict(val_aspy[2])  # extract the actual structure recipe
        val_type["value"] = val_type["value"][1:]  # change the value type to a scalar
        val_type = list(val_type.items())  # back to a list

        return val_type

    def _value_without_value(self, arrayval: Value, index: int | None = None) -> dict[str, Any]:
        # It would be straightforward to use arrayval.todict() but the value
        # could potentially be very large. So we use a more indirect way of
        # constructing it by iterating through the keys
        val_keys: list = arrayval.keys()
        val_keys.remove("value")

        val_dict = {}
        for val_key in val_keys:
            val_dict[val_key] = arrayval.todict(val_key)

        # We don't always have a value if changes are being made to other parts of
        # the structure, e.g. control limits

        if index and "value" in arrayval and len(arrayval["value"]) >= index:
            val_dict["value"] = arrayval["value"][index]
        else:
            # TODO: Default value isn't 0 for strings!
            # val_dict["value"] = 0
            pass

        return val_dict

    def scalarise(self, arrayval: Value, index: int | None = None) -> Value:
        """
        Convert the NTScalarArray into an NTScalar with the value of the
        index element in the array. If no index is provided a default value
        will be set.
        """

        # Constuct the new scalar value. This will have everything marked as changed
        val_id = self._get_value_id(arrayval)
        val_type = self._change_array_type_to_scalar_type(arrayval)
        val_dict = self._value_without_value(arrayval, index)
        value = Value(Type(val_type, id=val_id), val_dict)

        # Fix the changedSet so it matches that of the array passed in
        value.unmark()
        changed_set = arrayval.changedSet()
        for changed in changed_set:
            value.mark(changed)

        return value

    def _apply_gather(self, array_value: Value, scalar_value):
        if all(x in array_value.keys() for x in self._fields):
            overwrite_marked(array_value, scalar_value, self._fields)

    @check_applicable_init
    def init_rule(self, newpvstate: Value) -> RulesFlow:
        # Convert the new Value into scalar versions
        scalared_new_state = self.scalarise(newpvstate)

        gathered_value = self.scalarise(newpvstate)
        if isinstance(self._wrapped, BaseGatherableRule):
            self._wrapped.gather_init(gathered_value)

        # Loop through the array values applying the rules to each individual value
        newvals = []  # Use Ajit's trick to bypass the readonly value
        net_rule_flow = RulesFlow.CONTINUE
        for new_value in newpvstate["value"]:
            scalared_new_state["value"] = new_value

            rule_flow = self._wrapped.init_rule(scalared_new_state)
            if rule_flow == RulesFlow.ABORT:
                return RulesFlow.ABORT

            if rule_flow > net_rule_flow:  # Set the overall state to the worst we have encountered!
                net_rule_flow = rule_flow

            if isinstance(self._wrapped, BaseGatherableRule):
                self._wrapped.gather(scalared_new_state, gathered_value)

            newvals.append(scalared_new_state["value"])

        # Apply what was gathered
        newpvstate["value"] = newvals
        self._apply_gather(newpvstate, gathered_value)

        return net_rule_flow

    # NOTE: Performance will be terrible! Every rule and every value has to be iterated every time!
    # TODO: What's the correct behaviour if the new and old PV states have different lengths?
    # TODO: What is the correct behaviour for a Control Rule if the array size increases?
    # TODO: What if the Value["value"] has not changed?
    @check_applicable_post
    def post_rule(self, oldpvstate: Value, newpvstate: Value) -> RulesFlow:
        # Convert the current Value and new Value into scalar versions
        scalared_current_state = self.scalarise(oldpvstate)
        scalared_new_state = self.scalarise(newpvstate)

        gathered_value = self.scalarise(newpvstate)
        if isinstance(self._wrapped, BaseGatherableRule):
            self._wrapped.gather_init(gathered_value)

        # Loop through the array values applying the rules to each individual value
        newvals = []  # Use Ajit's trick to bypass the readonly value
        net_rule_flow = RulesFlow.CONTINUE
        for old_value, new_value in itertools.zip_longest(oldpvstate["value"], newpvstate["value"]):
            if old_value is not None:
                scalared_current_state["value"] = old_value
            else:
                scalared_current_state = None

            scalared_new_state["value"] = new_value

            rule_flow = self._wrapped.post_rule(scalared_current_state, scalared_new_state)

            if rule_flow == RulesFlow.ABORT:
                return RulesFlow.ABORT
            if rule_flow > net_rule_flow:  # Set the overall state to the worst we have encountered!
                net_rule_flow = rule_flow

            if isinstance(self._wrapped, BaseGatherableRule):
                self._wrapped.gather(scalared_new_state, gathered_value)

            newvals.append(scalared_new_state["value"])

        # Apply what was gathered
        newpvstate["value"] = newvals
        self._apply_gather(newpvstate, gathered_value)

        return net_rule_flow



import logging
from unittest.mock import patch

import numpy
import pytest
from p4p.nt import NTScalar

from p4pillon.definitions import AlarmSeverity
from p4pillon.rules import ControlRule, RulesFlow, ScalarToArrayWrapperRule, TimestampRule, ValueAlarmRule
from p4pillon.utils import overwrite_unmarked


class TestTimestamp:
    @pytest.mark.parametrize(
        "nttype, val",
        [
            ("d", 0),
            ("i", 0),
            ("s", "0"),
            ("ad", [0.5, 1.1, 2.2]),
            ("ai", [0, 1, 2]),
            ("as", ["0", "a", "longerstring"]),
        ],
    )
    @patch("time.time", return_value=123.456)
    def test_timestamp(self, _, nttype, val):
        rule = TimestampRule()

        assert rule._name == "timestamp"

        nt = NTScalar(nttype)
        old_state = nt.wrap(val)
        new_state = nt.wrap(val)
        overwrite_unmarked(old_state, new_state)

        assert new_state.changed("timeStamp") is False

        result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE
        assert new_state.changed("timeStamp") is True
        assert new_state["timeStamp.secondsPastEpoch"] == 123
        assert new_state["timeStamp.nanoseconds"] == 456000000

    @pytest.mark.parametrize(
        "nttype, val",
        [
            ("d", 0),
            ("i", 0),
            ("s", "0"),
            ("ad", [0.5, 1.1, 2.2]),
            ("ai", [0, 1, 2]),
            ("as", ["0", "a", "longerstring"]),
        ],
    )
    @patch("time.time", return_value=123.456)
    def test_timestamp_in_put(self, _, nttype, val):
        rule = TimestampRule()

        assert rule._name == "timestamp"

        nt = NTScalar(nttype)
        old_state = nt.wrap(val)
        new_state = nt.wrap(val)

        new_state["timeStamp.secondsPastEpoch"] = 123
        new_state["timeStamp.nanoseconds"] = 456000000

        overwrite_unmarked(old_state, new_state)

        assert new_state.changed("timeStamp") is True

        result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE


class TestControl:
    @pytest.mark.parametrize(
        "nttype, val",
        [
            ("d", 0),
            ("i", 0),
            ("s", "0"),
            ("ad", [0.5, 1.1, 2.2]),
            ("ai", [0, 1, 2]),
            ("as", ["0", "a", "longerstring"]),
        ],
    )
    def test_control_not_set(self, nttype, val, caplog):
        rule = ControlRule()

        assert rule._name == "control"

        # control not present
        nt = NTScalar(nttype)
        old_state = nt.wrap(val)
        new_state = nt.wrap(val)
        overwrite_unmarked(old_state, new_state)

        with caplog.at_level(logging.DEBUG):
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE
        assert len(caplog.records) == 1
        assert "Rule control.post_rule is not applicable" in str(caplog.records[0].getMessage())

    @pytest.mark.parametrize(
        "nttype, new_value, expected_value",
        [
            ("d", -6, -5),
            ("d", -1, -1),
            ("d", 1, 1),
            ("d", 6, 5),
            ("i", -6, -5),
            ("i", -1, -1),
            ("i", 1, 1),
            ("i", 6, 5),
            ("ad", [-6, -6, -6], [-5, -5, -5]),
            ("ad", [-1, -1, -1], [-1, -1, -1]),
            ("ad", [1, 1, 1], [1, 1, 1]),
            ("ad", [6, 6, 6], [5, 5, 5]),
            ("ai", [-6, -6, -6], [-5, -5, -5]),
            ("ai", [-1, -1, -1], [-1, -1, -1]),
            ("ai", [1, 1, 1], [1, 1, 1]),
            ("ai", [6, 6, 6], [5, 5, 5]),
        ],
    )
    def test_control(self, nttype, new_value, expected_value, caplog):
        nt = NTScalar(nttype, control=True)
        control_limits = {"limitLow": -5, "limitHigh": 5, "minStep": 1}
        if not nttype.startswith("a"):
            rule = ControlRule()
            old_state = nt.wrap({"value": 0.0, "control": control_limits})
        else:
            rule = ScalarToArrayWrapperRule(ControlRule())
            old_state = nt.wrap({"value": [0.0, 0.0, 0.0], "control": control_limits})

        new_state = nt.wrap({"value": new_value, "control": control_limits})
        overwrite_unmarked(old_state, new_state)

        with caplog.at_level(logging.DEBUG):
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == expected_value

            if new_value != expected_value:
                assert len(caplog.records) == 3
                assert f"control limit exceeded, changing value to {str(expected_value)}" in str(
                    caplog.records[2].getMessage()
                )
        else:
            numpy.testing.assert_array_equal(new_state["value"], expected_value)

    @pytest.mark.parametrize(
        "nttype, new_value, expected_value, expected_log, expected_log_index",
        [
            ("d", 2, 2, "", 1),
            ("d", 1, 0, "minStep", 1),
            ("d", 6, 5, "control limit exceeded", 2),
            ("i", 2, 2, "", 1),
            ("i", 1, 0, "minStep", 1),
            ("i", 6, 5, "control limit exceeded", 2),
            ("ad", [2, 2, 2], [2, 2, 2], ["", "", ""], 1),
            ("ad", [1, 1, 1], [0, 0, 0], ["minStep", "minStep", "minStep"], 3),
            (
                "ad",
                [6, 6, 6],
                [5, 5, 5],
                ["control limit exceeded", "control limit exceeded", "control limit exceeded"],
                2,
            ),
            ("ai", [2, 2, 2], [2, 2, 2], ["", "", ""], 1),
            ("ai", [1, 1, 1], [0, 0, 0], ["minStep", "minStep", "minStep"], 3),
            (
                "ai",
                [6, 6, 6],
                [5, 5, 5],
                ["control limit exceeded", "control limit exceeded", "control limit exceeded"],
                2,
            ),
        ],
    )
    def test_control_min_step(self, nttype, new_value, expected_value, expected_log, expected_log_index, caplog):
        nt = NTScalar(nttype, control=True)
        control_limits = {"limitLow": -5, "limitHigh": 5, "minStep": 2}
        if not nttype.startswith("a"):
            rule = ControlRule()
            old_state = nt.wrap({"value": 0.0, "control": control_limits})
        else:
            rule = ScalarToArrayWrapperRule(ControlRule())
            old_state = nt.wrap({"value": [0.0, 0.0, 0.0], "control": control_limits})
        new_state = nt.wrap({"value": new_value, "control": control_limits})
        overwrite_unmarked(old_state, new_state)

        with caplog.at_level(logging.DEBUG):
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == expected_value

            if new_value != expected_value:
                assert len(caplog.records) == 3
                assert expected_log in str(caplog.records[expected_log_index].getMessage())

        else:
            numpy.testing.assert_array_equal(new_state["value"], expected_value)

            if not numpy.array_equal(new_value, expected_value):
                assert len(caplog.records) == 9
                for item in zip(expected_log, caplog.records[expected_log_index:3:]):
                    assert item[0] in item[1].getMessage()

    @pytest.mark.parametrize(
        "nttype, control_changes, expected_value, read_only",
        [
            ("d", [-6, -5, 5, 2], -5, True),
            ("d", [-6, -5, 5, 2], -5, False),
            ("d", [-7, -10, 5, 2], -5, True),
            ("d", [-7, -10, 5, 2], -7, False),
        ],
    )
    def test_control_change_with_put(self, nttype, control_changes, expected_value, read_only):
        nt = NTScalar(nttype, control=True)
        control_limits = {"limitLow": -5, "limitHigh": 5, "minStep": 2}
        if not nttype.startswith("a"):
            rule = ControlRule()
            old_state = nt.wrap({"value": 0.0, "control": control_limits})
        else:
            rule = ScalarToArrayWrapperRule(ControlRule())
            old_state = nt.wrap({"value": [0.0, 0.0, 0.0], "control": control_limits})
        rule.read_only = read_only

        control_limits = {
            "limitLow": control_changes[1],
            "limitHigh": control_changes[2],
            "minStep": control_changes[3],
        }
        new_state = nt.wrap({"value": control_changes[0], "control": control_limits})
        overwrite_unmarked(old_state, new_state)

        with (
            patch("p4p.server.raw.SharedPV", autospec=True) as sharedpv,
            patch("p4p.server.ServerOperation", autospec=True) as server_op,
        ):
            sharedpv.current.return_value = nt.unwrap(old_state)
            server_op.value.return_value = nt.unwrap(new_state)
            result = rule.put_rule(sharedpv, server_op)  # New rules no long auto-call post_rule
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == expected_value
        else:
            numpy.testing.assert_array_equal(new_state["value"], expected_value)


class TestAlarmLimit:
    @pytest.mark.parametrize(
        "nttype, new_val, expected_severity, expected_message",
        [
            ("d", -10, AlarmSeverity.MAJOR_ALARM.value, "lowAlarm"),
            ("d", -5, AlarmSeverity.MINOR_ALARM.value, "lowWarning"),
            ("d", 0, AlarmSeverity.NO_ALARM.value, ""),
            ("d", 5, AlarmSeverity.MINOR_ALARM.value, "highWarning"),
            ("d", 10, AlarmSeverity.MAJOR_ALARM.value, "highAlarm"),
            ("i", -10, AlarmSeverity.MAJOR_ALARM.value, "lowAlarm"),
            ("i", -5, AlarmSeverity.MINOR_ALARM.value, "lowWarning"),
            ("i", 0, AlarmSeverity.NO_ALARM.value, ""),
            ("i", 5, AlarmSeverity.MINOR_ALARM.value, "highWarning"),
            ("i", 10, AlarmSeverity.MAJOR_ALARM.value, "highAlarm"),
            ("ad", [0, 0, -10], AlarmSeverity.MAJOR_ALARM.value, "lowAlarm"),
            ("ad", [0, -10, 0], AlarmSeverity.MAJOR_ALARM.value, "lowAlarm"),
            ("ad", [0, -5, 0], AlarmSeverity.MINOR_ALARM.value, "lowWarning"),
            ("ad", [0, 0, 0], AlarmSeverity.NO_ALARM.value, ""),
            ("ad", [0, 5, 0], AlarmSeverity.MINOR_ALARM.value, "highWarning"),
            ("ad", [0, 10, 0], AlarmSeverity.MAJOR_ALARM.value, "highAlarm"),
            ("ai", [0, 0, -10], AlarmSeverity.MAJOR_ALARM.value, "lowAlarm"),
            ("ai", [0, -10, 0], AlarmSeverity.MAJOR_ALARM.value, "lowAlarm"),
            ("ai", [0, -5, 0], AlarmSeverity.MINOR_ALARM.value, "lowWarning"),
            ("ai", [0, 0, 0], AlarmSeverity.NO_ALARM.value, ""),
            ("ai", [0, 5, 0], AlarmSeverity.MINOR_ALARM.value, "highWarning"),
            ("ai", [0, 10, 0], AlarmSeverity.MAJOR_ALARM.value, "highAlarm"),
        ],
    )
    def test_alarm_limits_value_change(self, nttype, new_val, expected_severity, expected_message, caplog):
        nt = NTScalar(nttype, valueAlarm=True)
        alarm_limits = {
            "active": True,
            "lowAlarmLimit": -9,
            "lowWarningLimit": -4,
            "highWarningLimit": 4,
            "highAlarmLimit": 9,
            "lowAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
            "lowWarningSeverity": AlarmSeverity.MINOR_ALARM.value,
            "highAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
            "highWarningSeverity": AlarmSeverity.MINOR_ALARM.value,
        }
        if not nttype.startswith("a"):
            rule = ValueAlarmRule()
            old_state = nt.wrap({"value": 0.0, "valueAlarm": alarm_limits})
        else:
            rule = ScalarToArrayWrapperRule(ValueAlarmRule())
            old_state = nt.wrap({"value": [0.0, 0.0, 0.0], "valueAlarm": alarm_limits})
        new_state = nt.wrap({"value": new_val, "valueAlarm": alarm_limits})
        overwrite_unmarked(old_state, new_state)

        with caplog.at_level(logging.DEBUG):
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == new_val
        else:
            numpy.testing.assert_array_equal(new_state["value"], new_val)

        assert new_state["alarm.severity"] == expected_severity
        assert new_state["alarm.message"] == expected_message

    @pytest.mark.parametrize(
        "nttype, new_val", [("d", -10), ("i", -10), ("ad", [-10, -10, -10]), ("ai", [-10, -10, -10])]
    )
    def test_alarm_limits_not_active(self, nttype, new_val, caplog):
        nt = NTScalar(nttype, valueAlarm=True)
        alarm_limits = {
            "active": False,
            "lowAlarmLimit": -9,
            "lowAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
        }
        if not nttype.startswith("a"):
            rule = ValueAlarmRule()
            old_state = nt.wrap({"value": 0.0, "valueAlarm": alarm_limits})
        else:
            rule = ScalarToArrayWrapperRule(ValueAlarmRule())
            old_state = nt.wrap({"value": [0.0, 0.0, 0.0], "valueAlarm": alarm_limits})
        new_state = nt.wrap({"value": new_val, "valueAlarm": alarm_limits})
        overwrite_unmarked(old_state, new_state)

        with caplog.at_level(logging.DEBUG):
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == new_val
        else:
            numpy.testing.assert_array_equal(new_state["value"], new_val)

        assert new_state["alarm.severity"] == AlarmSeverity.NO_ALARM.value
        assert new_state["alarm.message"] == ""

    @pytest.mark.parametrize(
        "nttype, new_val", [("d", -10), ("i", -10), ("ad", [-10, -10, -10]), ("ai", [-10, -10, -10])]
    )
    def test_alarm_limits_not_present(self, nttype, new_val, caplog):
        nt = NTScalar(nttype)
        if not nttype.startswith("a"):
            rule = ValueAlarmRule()
            old_state = nt.wrap({"value": 0.0})
        else:
            rule = ScalarToArrayWrapperRule(ValueAlarmRule())
            old_state = nt.wrap({"value": [0.0, 0.0, 0.0]})
        new_state = nt.wrap({"value": new_val})
        overwrite_unmarked(old_state, new_state)

        with caplog.at_level(logging.DEBUG):
            result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == new_val
        else:
            numpy.testing.assert_array_equal(new_state["value"], new_val)

        assert new_state["alarm.severity"] == AlarmSeverity.NO_ALARM.value
        assert new_state["alarm.message"] == ""

    @pytest.mark.parametrize(
        "nttype, new_val", [("d", -10), ("i", -10), ("ad", [-10, -10, -10]), ("ai", [-10, -10, -10])]
    )
    def test_alarm_limits_from_alarm_state_to_none(self, nttype, new_val):
        # here we make sure that changing the value from a previous alarm state will put us
        # a no alarm state

        nt = NTScalar(nttype, valueAlarm=True)
        alarm_limits = {
            "active": True,
            "lowAlarmLimit": -9,
            "lowWarningLimit": -4,
            "highWarningLimit": 4,
            "highAlarmLimit": 9,
            "lowAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
            "lowWarningSeverity": AlarmSeverity.MINOR_ALARM.value,
            "highAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
            "highWarningSeverity": AlarmSeverity.MINOR_ALARM.value,
        }
        old_state = nt.wrap(
            {
                "value": new_val,
                "valueAlarm": alarm_limits,
                "alarm": {
                    "severity": AlarmSeverity.MAJOR_ALARM.value,
                    "message": "highAlarm",
                    "status": 0,
                },
            }
        )
        if not nttype.startswith("a"):
            rule = ValueAlarmRule()
            new_state = nt.wrap({"value": 0})
        else:
            rule = ScalarToArrayWrapperRule(ValueAlarmRule())
            new_state = nt.wrap({"value": [0, 0, 0]})

        overwrite_unmarked(old_state, new_state)

        result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == 0
        else:
            numpy.testing.assert_array_equal(new_state["value"], [0, 0, 0])

        assert new_state["alarm.severity"] == AlarmSeverity.NO_ALARM.value
        assert new_state["alarm.message"] == ""

    @pytest.mark.parametrize(
        "nttype, limit_change, new_limit, expected_severity, expected_message",
        [
            ("d", "lowAlarmLimit", 2, AlarmSeverity.MAJOR_ALARM, "lowAlarm"),
            ("d", "lowWarningLimit", 2, AlarmSeverity.MINOR_ALARM, "lowWarning"),
            ("d", "highWarningLimit", 0, AlarmSeverity.MINOR_ALARM, "highWarning"),
            ("d", "highAlarmLimit", 0, AlarmSeverity.MAJOR_ALARM, "highAlarm"),
            ("i", "lowAlarmLimit", 2, AlarmSeverity.MAJOR_ALARM, "lowAlarm"),
            ("i", "lowWarningLimit", 2, AlarmSeverity.MINOR_ALARM, "lowWarning"),
            ("i", "highWarningLimit", 0, AlarmSeverity.MINOR_ALARM, "highWarning"),
            ("i", "highAlarmLimit", 1, AlarmSeverity.MAJOR_ALARM, "highAlarm"),
            ("ad", "lowAlarmLimit", 2, AlarmSeverity.MAJOR_ALARM, "lowAlarm"),
            ("ad", "lowWarningLimit", 2, AlarmSeverity.MINOR_ALARM, "lowWarning"),
            ("ad", "highWarningLimit", 0, AlarmSeverity.MINOR_ALARM, "highWarning"),
            ("ad", "highAlarmLimit", 0, AlarmSeverity.MAJOR_ALARM, "highAlarm"),
            ("ai", "lowAlarmLimit", 2, AlarmSeverity.MAJOR_ALARM, "lowAlarm"),
            ("ai", "lowWarningLimit", 2, AlarmSeverity.MINOR_ALARM, "lowWarning"),
            ("ai", "highWarningLimit", 0, AlarmSeverity.MINOR_ALARM, "highWarning"),
            ("ai", "highAlarmLimit", 1, AlarmSeverity.MAJOR_ALARM, "highAlarm"),
        ],
    )
    def test_alarm_limits_changing_limits(self, nttype, limit_change, new_limit, expected_severity, expected_message):
        # if we change the limit on an alarm, we want to make sure that the new alarm state
        # is calculated based on the new limits
        if not nttype.startswith("a"):
            rule = ValueAlarmRule()
            old_value = 0
            new_value = 1
        else:
            rule = ScalarToArrayWrapperRule(ValueAlarmRule())
            old_value = [0, 0, 0]
            new_value = [0, 1, 0]

        nt = NTScalar(nttype, valueAlarm=True)
        alarm_limits = {
            "active": True,
            "lowAlarmLimit": -9,
            "lowWarningLimit": -4,
            "highWarningLimit": 4,
            "highAlarmLimit": 9,
            "lowAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
            "lowWarningSeverity": AlarmSeverity.MINOR_ALARM.value,
            "highAlarmSeverity": AlarmSeverity.MAJOR_ALARM.value,
            "highWarningSeverity": AlarmSeverity.MINOR_ALARM.value,
        }
        old_state = nt.wrap(
            {
                "value": old_value,
                "valueAlarm": alarm_limits,
            }
        )
        new_state = nt.wrap(
            {
                "value": new_value,
                f"valueAlarm.{limit_change}": new_limit,
            }
        )
        overwrite_unmarked(old_state, new_state)

        result = rule.post_rule(old_state, new_state)

        assert result is RulesFlow.CONTINUE

        if not nttype.startswith("a"):
            assert new_state["value"] == new_value
        else:
            numpy.testing.assert_array_equal(new_state["value"], new_value)

        assert new_state["alarm.severity"] == expected_severity.value
        assert new_state["alarm.message"] == expected_message



from p4p.nt import NTScalar
from p4p.server import Server

from p4pillon.thread.sharednt import SharedNT

pv = SharedNT(
    nt=NTScalar(
        "d",
        valueAlarm=True,
    ),  # scalar double
    initial={"value": 4.5, "valueAlarm.active": True, "valueAlarm.highAlarmLimit": 17},
)  # setting initial value also open()'s


Server.forever(
    providers=[
        {
            "demo:pv:name": pv,  # PV name only appears here
        }
    ]
)  # runs until KeyboardInterrupt



"""Concrete implementations of PVRecipe using threads"""

from __future__ import annotations

from p4pillon.definitions import PVTypes
from p4pillon.pvrecipe import BasePVRecipe
from p4pillon.pvrecipe import PVScalarRecipe as _PVScalarRecipe
from p4pillon.server.thread import SharedPV


class PVScalarRecipe(_PVScalarRecipe):
    """
    A recipe for creating a PV of type NTScalar.

    This class is used to create a PV that represents a scalar value,
    allowing for the definition of initial values, descriptions, and other properties.
    """

    def create_pv(self, pv_name: str | None = None) -> SharedPV:
        """Turn the recipe into an actual NTScalar, NTEnum, or
        other BasePV derived object"""

        self._config_display()
        self._config_control()
        self._config_alarm_limit()

        return super().build_pv()


class PVScalarArrayRecipe(_PVScalarRecipe):
    """
    A recipe for creating a PV of type NTScalar.

    This class is used to create a PV that represents an array of scalar values,
    allowing for the definition of initial values, descriptions, and other properties.
    """

    def create_pv(self, pv_name: str | None = None) -> SharedPV:
        """Turn the recipe into an actual NTScalar with an array"""

        self._config_display()
        self._config_control()
        self._config_alarm_limit()

        if not isinstance(self.initial_value, list):
            self.initial_value = [self.initial_value]

        return super().build_pv()


class PVEnumRecipe(BasePVRecipe):
    """
    A recipe for creating a PV of type NTEnum.

    This class is used to create a PV that represents an enumeration type,
    allowing for the definition of enum values and their corresponding labels.
    """

    def __post_init__(self):
        super().__post_init__()
        if not self.pvtype == PVTypes.ENUM:
            raise ValueError(f"Unsupported pv type {self.pvtype} for class {{self.__class__.__name__}}")

    def create_pv(self, pv_name: str | None = None) -> SharedPV:
        """Turn the recipe into an actual NTEnum"""

        return super().build_pv()



from unittest.mock import MagicMock, patch

import pytest


@pytest.fixture
def mock_recipe():
    pv = MagicMock()
    return pv


@pytest.fixture
def mock_server_op():
    with patch("p4pillon.handlers.ServerOperation", autospec=True) as server_op:
        yield server_op


@pytest.fixture
def mock_server():
    with patch("p4pillon.server.Server", autospec=True) as server:
        yield server


@pytest.fixture
def mock_ntpv():
    with patch("p4pillon.thread.sharednt.SharedNT", autospec=True) as shared_pv:
        yield shared_pv


@pytest.fixture
def mock_provider():
    with patch("p4p.server.StaticProvider", autospec=True) as provider:
        yield provider



"""Read configuration from a YAML file"""

from __future__ import annotations

import logging
from typing import Any

import yaml

from p4pillon.definitions import PVTypes
from p4pillon.pvrecipe import BasePVRecipe
from p4pillon.server.simpleserver import BaseSimpleServer
from p4pillon.thread.pvrecipe import PVEnumRecipe, PVScalarArrayRecipe, PVScalarRecipe

logger = logging.getLogger(__name__)


def parse_config_file(filename: str, server: BaseSimpleServer | None = None) -> dict[str, BasePVRecipe]:
    """
    Parse a yaml file and return a dictionary of PVScalarRecipe objects.
    Optionally add the pvs to a server if server != None
    """
    pvconfigs = {}
    with open(filename, encoding="utf8") as f:
        pvconfigs = yaml.load(f, yaml.SafeLoader)

    return parse_config(pvconfigs, server)


def parse_config_string(yaml_str: str, server: BaseSimpleServer | None = None) -> dict[str, BasePVRecipe]:
    """
    Parse a yaml string and return a dictionary of PVScalarRecipe objects.
    Optionally add the pvs to a server if server != None
    """
    pvconfigs = {}
    pvconfigs = yaml.load(yaml_str, yaml.SafeLoader)

    return parse_config(pvconfigs, server)


def parse_config(
    yaml_obj: dict[str, dict[str, Any]], server: BaseSimpleServer | None = None
) -> dict[str, BasePVRecipe]:
    """
    Parse a dictionary that has been filled using yaml.load() and return a dictionary of PVScalarRecipe objects.
    Optionally add the pvs to a server if server != None
    """

    pvrecipes = {}

    logger.debug("Processing yaml: \n%r", yaml_obj)

    for name, config in yaml_obj.items():
        recipe = process_config(name, config)
        pvrecipes[name] = recipe

        if server is not None:
            server.add_pv(name, recipe)

    return pvrecipes


def process_config(pvname: str, pvdetails: dict[str, Any]) -> BasePVRecipe:
    """
    Process the configuration of a single PV and update pvrecipe accordingly.

    The configuration is a dictionary, currently constructed from a YAML file.
    An example:

        pvname:
           type: 'd'
           valueAlarm: False

    The returned values are in three parts:
        construct_settings - passed to the default constructor of SharedPV

        initial_value   - a value initial value if one has not been supplied,
        SharedPV won't open() the PV unless this is provided

        config_settings - SharedPV only supports setting some values in the
        constructor, the rest need to be added in a later post()
    """

    logger.debug("Processing configuration for pv %s, config is %r", pvname, pvdetails)

    # Check that type and description are specified, absence is a syntax error
    if "type" not in pvdetails:
        raise SyntaxError(f"'type' not specified in record {pvname}")
    if "description" not in pvdetails:
        raise SyntaxError(f"'description' not specified in record {pvname}")

    initial = pvdetails.get("initial")
    array_size = pvdetails.get("array_size", 1)
    pvtype = pvdetails["type"]
    if not initial:
        # If it's a number set it to 0, if it's a string make it empty
        # If it's something else an initial value needs to be supplied
        if pvtype == "DOUBLE" or pvtype == "INTEGER":
            if array_size > 1:
                initial = [0] * array_size
            else:
                initial = 0
        elif pvtype == "STRING":
            if array_size > 1:
                initial = [""] * array_size
            else:
                initial = ""
        else:
            raise SyntaxError(f"for PV {pvname} of type '{pvtype}' an initial value must be supplied")

    if isinstance(initial, list):
        pvrecipe = PVScalarArrayRecipe(PVTypes[pvdetails["type"]], pvdetails["description"], initial)
    elif pvtype == "ENUM":
        pvrecipe = PVEnumRecipe(PVTypes[pvdetails["type"]], pvdetails["description"], initial)
    else:
        pvrecipe = PVScalarRecipe(PVTypes[pvdetails["type"]], pvdetails["description"], initial)

    supported_configs = [("read_only", bool), ("calc", dict)]
    for config, config_type in supported_configs:
        # Process variables in the configuration that are attributes of the pvrecipe class
        temp_config = pvdetails.get(config)
        if temp_config is not None and isinstance(temp_config, config_type):
            setattr(pvrecipe, config, temp_config)

    if "control" in pvdetails:
        pvrecipe.set_control_limits(**get_field_config(pvdetails, "control"))
    if "display" in pvdetails:
        pvrecipe.set_display_limits(**get_field_config(pvdetails, "display"))
    if "valueAlarm" in pvdetails:
        pvrecipe.set_alarm_limits(**get_field_config(pvdetails, "valueAlarm"))

    return pvrecipe


def get_field_config(pvdetails: dict, field_name: str) -> dict:
    """Get a specified field from the configuration"""
    config = pvdetails.get(field_name)
    if config is None:
        config = {}
    return config



from collections import OrderedDict

import pytest

from p4pillon.nt import NTEnum, NTScalar
from p4pillon.server.raw import Handler
from p4pillon.thread.sharednt import SharedNT


@pytest.mark.parametrize(
    "pvtype, expected_handlername",
    [
        ("d", ["control", "alarm", "alarm_limit", "timestamp"]),
        ("ad", ["control", "alarm", "alarm_limit", "timestamp"]),
        ("i", ["control", "alarm", "alarm_limit", "timestamp"]),
        ("ai", ["control", "alarm", "alarm_limit", "timestamp"]),
    ],
)
def testntscalar_create(pvtype, expected_handlername):
    testpv = SharedNT(
        nt=NTScalar(pvtype),
    )

    assert len(testpv.handler) == 4
    assert list(testpv.handler.keys()) == expected_handlername


@pytest.mark.parametrize(
    "pvtype, expected_handlername",
    [
        (
            "d",
            [
                "control",
                "alarm",
                "alarm_limit",
            ],
        ),
        (
            "ad",
            [
                "control",
                "alarm",
                "alarm_limit",
            ],
        ),
        (
            "i",
            [
                "control",
                "alarm",
                "alarm_limit",
            ],
        ),
        (
            "ai",
            [
                "control",
                "alarm",
                "alarm_limit",
            ],
        ),
    ],
)
def testntscalar_create_with_handlers(pvtype, expected_handlername):
    testpv = SharedNT(
        nt=NTScalar(pvtype),
        auth_handlers=OrderedDict({"pre1": Handler(), "pre2": Handler()}),
        user_handlers=OrderedDict({"post1": Handler(), "post2": Handler()}),
    )

    assert len(testpv.handler) == 8
    assert list(testpv.handler.keys()) == ["pre1", "pre2", *expected_handlername, "post1", "post2", "timestamp"]


def testntenum_create():
    testpv = SharedNT(nt=NTEnum(), initial={"index": 0, "choices": ["OFF", "ON"]})

    assert len(testpv.handler) == 1
    assert list(testpv.handler.keys()) == ["timestamp"]


def testntenum_create_with_handlers():
    testpv = SharedNT(
        nt=NTEnum(),
        initial={"index": 0, "choices": ["OFF", "ON"]},
        auth_handlers=OrderedDict({"pre1": Handler(), "pre2": Handler()}),
        user_handlers=OrderedDict({"post1": Handler(), "post2": Handler()}),
    )

    assert len(testpv.handler) == 5
    assert list(testpv.handler.keys()) == ["pre1", "pre2", "post1", "post2", "timestamp"]


@pytest.mark.filterwarnings("ignore")  # Ignore "RuntimeError: Empty SharedPV" warning
def testbadnt():
    with pytest.raises(NotImplementedError):
        SharedNT(nt=bool)



"""
Monkey patch in required changes to NTEnum
"""

from p4p.nt import *  # type: ignore # noqa: F403
from p4p.nt import NTBase, NTEnum  # noqa: F401

from p4pillon.nt.enum import NTEnum as _NTEnum

NTEnum = _NTEnum  # noqa: F811



"""
An example of using pvrecipe with asyncio to create an NTScalar PV.
The created PV will be called 'demo:pv:name' and will have a default value of 17.5.
It will also have a description and alarm thresholds. If the value is changed, the
alarm secverity will automatically update.
"""

import asyncio
import logging

from p4p.server import Server, StaticProvider

from p4pillon.asyncio.pvrecipe import PVScalarRecipe
from p4pillon.definitions import PVTypes

DEFAULT_TIMEOUT = 1


class AsyncProviderWrapper:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop
        self._provider = StaticProvider()

        self._loop.run_until_complete(asyncio.wait_for(self.asyncSetUp(), DEFAULT_TIMEOUT))

    @property
    def providers(self) -> tuple[StaticProvider]:
        return (self._provider,)

    async def asyncSetUp(self):
        logging.info("Async set up.")

        pvrecipe_double1 = PVScalarRecipe(PVTypes.DOUBLE, "An example double PV", 5.0)
        pvrecipe_double1.initial_value = 17.5
        pvrecipe_double1.description = "A different default value for the PV"
        # try setting a different value for the timestamp
        pvrecipe_double1.set_timestamp(1729699237.8525229)
        pvrecipe_double1.set_alarm_limits(low_warning=2, high_alarm=9)

        pv_double1 = pvrecipe_double1.create_pv()

        self._provider.add("demo:pv:name", pv_double1)


def main():
    loop = asyncio.new_event_loop()
    provider_wrapper = AsyncProviderWrapper(loop)

    try:
        # `Server.forever()` is for p4p threading and shouldn't
        # be used with async.
        server = Server(provider_wrapper.providers)
        with server:
            done = asyncio.Event()

            # loop.add_signal_handler(signal.SIGINT, done.set)
            # loop.add_signal_handler(signal.SIGTERM, done.set)
            loop.run_until_complete(done.wait())
    finally:
        loop.close()


if __name__ == "__main__":
    main()



import logging
from pathlib import Path
from unittest.mock import patch

import pytest
from p4p.server import StaticProvider

from p4pillon.thread.server import SimpleServer

root_dir = Path(__file__).parents[2]


def test_server_instantiation():
    server = SimpleServer(
        prefix="DEV:",
    )
    assert server.prefix == "DEV:"

    # before we explicitly call `start()`, the server shouldn't exist
    assert server._server is None
    assert isinstance(server._provider, StaticProvider)
    # and we also shouldn't have any PVs configured
    assert server.pvlist == []


@pytest.mark.parametrize(
    "pv_name",
    [("TEST:PV"), ("DEV:TEST:PV")],
)
def test_server_retrieve_pvs(mock_recipe, pv_name):
    server = SimpleServer(
        prefix="DEV:",
    )
    server.add_pv(pv_name, mock_recipe)

    # we should be able to access the PV either with the full prefix added or without it
    assert server["TEST:PV"] == mock_recipe.create_pv.return_value
    assert server["DEV:TEST:PV"] == mock_recipe.create_pv.return_value


@patch("p4pillon.thread.server.StaticProvider", autospec=True)
@patch("p4pillon.thread.server.Server", autospec=True)
def test_server_start(server, provider, caplog, mock_ntpv):
    test_server = SimpleServer(
        prefix="DEV:",
    )

    mock_ntpv.on_start_methods = []
    test_server._pvs = {"DEV:TEST:PV:1": mock_ntpv}
    print(len(mock_ntpv.on_start_methods))

    assert test_server._running is False
    with caplog.at_level(logging.DEBUG):
        test_server.start()
    assert len(caplog.records) == 1
    provider.return_value.add.assert_called_once_with("DEV:TEST:PV:1", mock_ntpv)
    server.assert_called_once_with(providers=[provider.return_value])
    assert test_server._running is True


@patch("p4pillon.thread.server.StaticProvider", autospec=True)
@patch("p4pillon.thread.server.Server", autospec=True)
@patch("p4pillon.pvrecipe.PVScalarRecipe", autospec=True)
def test_server_add_pv(recipe, server, provider, caplog):
    test_server = SimpleServer(
        prefix="DEV:",
    )

    test_server.start()

    with caplog.at_level(logging.DEBUG):
        new_name = "TEST:PV:2"
        test_server.add_pv(new_name, recipe.return_value)
    assert test_server["TEST:PV:2"] is recipe.return_value.create_pv.return_value
    provider.return_value.add.assert_called_once_with("DEV:TEST:PV:2", recipe.return_value.create_pv.return_value)

    assert len(caplog.messages) == 1
    assert caplog.messages[0] == "Added DEV:TEST:PV:2 to server"


@patch("p4pillon.thread.server.StaticProvider", autospec=True)
@patch("p4pillon.thread.server.Server", autospec=True)
def test_server_stop(server, provider, caplog, mock_ntpv):
    test_server = SimpleServer(
        prefix="DEV:",
    )

    test_server._running = True
    test_server._server = server.return_value
    test_server._pvs = {"DEV:TEST:PV:1": mock_ntpv}

    with caplog.at_level(logging.DEBUG):
        test_server.stop()

    mock_ntpv.close.assert_called_once_with()
    provider.return_value.remove.assert_called_once_with("DEV:TEST:PV:1")
    server.return_value.stop.assert_called_once_with()
    assert test_server._running is False


@patch("p4pillon.thread.server.StaticProvider", autospec=True)
@patch("p4pillon.thread.server.Server", autospec=True)
@patch("p4pillon.pvrecipe.PVScalarRecipe", autospec=True)
def test_server_remove_pv(recipe, server, provider, caplog, mock_ntpv):
    test_server = SimpleServer(
        prefix="DEV:",
    )

    test_server._pvs["DEV:TEST:PV:1"] = mock_ntpv
    test_server._running = True
    test_server._server = server.return_value

    with caplog.at_level(logging.DEBUG):
        test_server.remove_pv("TEST:PV:1")

    mock_ntpv.close.assert_called_once_with()
    provider.return_value.remove.assert_called_once_with("DEV:TEST:PV:1")

    assert test_server._pvs.get("DEV:TEST:PV:1") is None

    assert len(caplog.messages) == 1
    assert caplog.messages[0] == "Removed DEV:TEST:PV:1 from server"



import time
from typing import Any

from p4p.client.thread import Context


def put_different_value_scalar(ctx: Context, pvname: str) -> tuple[str | Any, float]:
    """
    Change the value of a process variable (PV) to ensure it is different from its current value.

    Parameters:
    -----------
    ctx : Context
        The context object that provides methods to get and put the value of the PV.
    pvname : str
        The name of the PV whose value is to be changed.

    Returns:
    --------
    tuple
        A tuple containing the new value of the PV and the Unix timestamp when the update was made.

    Example:
    --------
    >>> ctx = Context()
    >>> pvname = "temperature_sensor_1"
    >>> new_value, timestamp = put_different_value(ctx, pvname)
    >>> print(f"New value: {new_value}, Updated at: {timestamp}")
    """
    current_val = ctx.get(pvname).raw.todict()["value"]
    if isinstance(current_val, str):
        put_val = current_val + "1"
    else:
        put_val = current_val + 1
    put_timestamp = time.time()
    ctx.put(pvname, put_val)
    time.sleep(0.1)
    return put_val, put_timestamp


def put_metadata(ctx: Context, pvname: str, field: str, value):
    """
    Update the metadata of a process variable in the given context and return the timestamp of the update.

    Parameters:
    -----------
    ctx : Context
        The context object that provides the method to update the PV.
    pvname : str
        The name of the PV whose metadata is to be updated.
    field : str
        The specific metadata field that needs to be updated. For subfields use dot notation e.g. valueAlarm.highAlarmLimit
    value
        The value to set for the specified metadata field. The type of this value can vary based on the field.

    Returns:
    --------
    float
        The Unix timestamp when the metadata was updated.

    Example:
    --------
    >>> ctx = Context()
    >>> pvname = "temperature_sensor_1"
    >>> field = "display.units"
    >>> value = "Celsius"
    >>> timestamp = put_metadata(ctx, pvname, field, value)
    >>> print(f"Metadata updated at {timestamp}")
    """
    put_timestamp = time.time()
    ctx.put(
        pvname,
        {field: value},
    )
    time.sleep(0.1)
    return put_timestamp


def put_different_value_enum(ctx: Context, pvname: str) -> tuple[dict | Any, float]:
    """
    Change the value of an Enum process variable (PV) to ensure it is different from its current value.

    Parameters:
    -----------
    ctx : Context
        The context object that provides methods to get and put the value of the PV.
    pvname : str
        The name of the PV whose value is to be changed.

    Returns:
    --------
    tuple
        A tuple containing the new value of the PV and the Unix timestamp when the update was made.

    Example:
    --------
    >>> ctx = Context()
    >>> pvname = "valve_status_1"
    >>> new_value, timestamp = put_different_value(ctx, pvname)
    >>> print(f"New value: {new_value}, Updated at: {timestamp}")
    """
    current_val = ctx.get(pvname).raw.todict()["value"]
    current_val["index"] = not (current_val["index"])
    # put_val = {"index": , "choices": current_val["choices"]}
    put_timestamp = time.time()
    ctx.put(pvname, current_val["index"])
    time.sleep(0.1)
    return current_val, put_timestamp



"""Utility function for p4pillon"""

from __future__ import annotations

from collections.abc import Callable
from typing import cast

from p4p import Value


def time_in_seconds_and_nanoseconds(timestamp: float) -> tuple[int, int]:
    """Convert a timestamp into separate integer seconds and nanoseconds"""
    seconds = int(timestamp // 1)
    nanoseconds = int((timestamp % 1) * 1e9)
    return seconds, nanoseconds


def recurse_values(value1: Value, value2: Value, func: Callable[[Value, Value, str], None], keys=None) -> bool:
    """Recurse through two Values with the same structure and apply a supplied to the leaf nodes"""
    if not keys:
        keys = cast(list[str], value1.keys())

    for key in keys:
        if isinstance(value1[key], Value) and isinstance(value2[key], Value):
            if not recurse_values(value1[key], value2[key], func):
                return False
        else:
            func(value1, value2, key)

    return True


def overwrite_marked(current: Value, update: Value, fields: list[str] | None = None) -> None:
    """
    Overwrite all of the unmarked fields in one Value with fields from another Value.

    This makes the changes in place rather than returning a copy.
    """

    def overwrite_changed_key(update_leaf: Value, current_leaf: Value, key: str) -> None:
        """
        Given a leaf node in the update Value tree, check whether it is changed and, if so,
        change the matching current leaf to its value
        """
        if update_leaf.changed(key):
            current_leaf[key] = update_leaf[key]

    if not fields:
        fields = cast(list[str], current.keys())

    recurse_values(update, current, overwrite_changed_key, fields)


def overwrite_unmarked(current: Value, update: Value, fields: list[str] | None = None) -> None:
    """
    Overwrite all of the unmarked fields in one Value with fields from another Value.

    This makes the changes in place rather than returning a copy.
    """

    def overwrite_unchanged_key(update_leaf: Value, current_leaf: Value, key: str) -> None:
        """
        Given a leaf node in the update Value tree, check whether it is unchanged and, if so,
        set it equal to the equivalent leaf node in the current Value tree. Then mark the new
        value for the leaf as unchanged.
        """
        if not update_leaf.changed(key):
            update_leaf[key] = current_leaf[key]
            update_leaf.mark(key, val=False)

    if not fields:
        fields = cast(list[str], current.keys())

    recurse_values(update, current, overwrite_unchanged_key, fields)



from p4p.nt import NTScalar

from p4pillon.server.thread import Handler, SharedPV


class TestThreadHandler:
    """
    Test Handler open(), post() and close() functions called correctly by SharedPV. Note that:
    - TestRPC, TestFirstLast already test onFirstConnect() and onLastDisconnect().
    - TestGPM, TestPVRequestMask already test put().
    - TestRPC, TestRPC2 already test rpc().
    """

    class HandlerTest(Handler):
        def __init__(self):
            self.last_op = "init"

        def open(self, value):
            self.last_op = "open"
            value["value"] = 17

        def post(self, pv, value):
            self.last_op = "post"
            value["value"] = value["value"] * 2

        def close(self, pv):
            self.last_op = "close"

    def setup_method(self, _method):
        self.handler = self.HandlerTest()
        self.pv = SharedPV(handler=self.handler, nt=NTScalar("d"))

    def test_open(self):
        # Setup sets the initial value to 5, but the Handler open() overrides
        self.pv.open(5)
        assert self.handler.last_op == "open"
        assert self.pv.current() == 17.0

    def test_post(self):
        self.pv.open(5)
        self.pv.post(13.0)
        assert self.handler.last_op == "post"
        assert self.pv.current() == 26.0

    def test_close(self):
        self.pv.open(5)
        self.pv.close(sync=True)
        assert self.handler.last_op == "close"

    def teardown_method(self, _method):
        self.pv.close()
        del self.handler
        del self.pv



"""Example of simplifed interface for NTScalar creation"""

from __future__ import annotations

import collections.abc
import dataclasses
import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar
from typing import SupportsFloat as Numeric  # Hack to type hint number types

from p4pillon.nt import NTEnum, NTScalar
from p4pillon.server.asyncio import SharedPV as SharedPV_asyncio
from p4pillon.server.thread import SharedPV as SharedPV_threaded
from p4pillon.thread.sharednt import SharedNT

from .definitions import (
    MAX_FLOAT,
    MAX_INT32,
    MIN_FLOAT,
    MIN_INT32,
    AlarmSeverity,
    Format,
    PVTypes,
)
from .utils import time_in_seconds_and_nanoseconds

NumericTypeT = TypeVar("NumericTypeT", int, Numeric)
SharedPvT = TypeVar("SharedPvT", SharedPV_threaded, SharedPV_asyncio)

logger = logging.getLogger(__name__)


@dataclass
class Timestamp:
    """Very simple timestamp class"""

    time: float

    def time_in_seconds_and_nanoseconds(self) -> tuple[int, int]:
        """Convert to EPICS style structured timestamp"""
        return time_in_seconds_and_nanoseconds(self.time)


@dataclass
class Control(Generic[NumericTypeT]):
    """Set limits on permitted values"""

    limit_low: NumericTypeT | None = None
    limit_high: NumericTypeT | None = None
    min_step: NumericTypeT = 0


@dataclass
class Display(Generic[NumericTypeT]):
    """Set limits on values that will be displayed"""

    limit_low: NumericTypeT | None = None
    limit_high: NumericTypeT | None = None
    units: str = ""
    format: Format = Format.DEFAULT
    precision: int = 2


@dataclass
class AlarmLimit(Generic[NumericTypeT]):
    """Conditions to test for alarms"""

    active: bool = True
    low_alarm_limit: NumericTypeT | None = None
    low_warning_limit: NumericTypeT | None = None
    high_warning_limit: NumericTypeT | None = None
    high_alarm_limit: NumericTypeT | None = None
    low_alarm_severity: AlarmSeverity = AlarmSeverity.MAJOR_ALARM
    low_warning_severity: AlarmSeverity = AlarmSeverity.MINOR_ALARM
    high_warning_severity: AlarmSeverity = AlarmSeverity.MINOR_ALARM
    high_alarm_severity: AlarmSeverity = AlarmSeverity.MAJOR_ALARM
    hysteresis: NumericTypeT = 0


@dataclass
class BasePVRecipe(Generic[SharedPvT], ABC):
    """A description of how to build a PV"""

    pvtype: PVTypes
    description: str
    initial_value: Numeric | list | str

    # Alarm: alarm = field(init=False)
    timestamp: Timestamp | None = None
    display: Display | None = None
    control: Control | None = None
    alarm_limit: AlarmLimit | None = None

    read_only: bool = False

    def __post_init__(self):
        """Anything that isn't done by the automatically created __init__"""

        # Initialise the members that the default init doesn't cover
        # Specifically these are the ones tagged with field(init=False)
        self.construct_settings = {}
        self.config_settings = {}

        self.construct_settings["valtype"] = self.pvtype.value
        self.construct_settings["extra"] = [("descriptor", "s")]
        self.config_settings["descriptor"] = self.description

    @abstractmethod
    def create_pv(self, pv_name: str | None = None) -> SharedPvT:
        """Turn the recipe into an NT object with an array"""

        raise NotImplementedError

    def _config_timestamp(self):
        if self.timestamp:
            seconds, nanoseconds = self.timestamp.time_in_seconds_and_nanoseconds()
        else:
            seconds, nanoseconds = Timestamp(time.time()).time_in_seconds_and_nanoseconds()
        self.config_settings["timeStamp.secondsPastEpoch"] = seconds
        self.config_settings["timeStamp.nanoseconds"] = nanoseconds

    def build_pv(
        self,
    ) -> SharedPvT:
        """
        This method is called by create_pv in the child classes after construct settings is set.
        """
        debug_str = (
            f"Building pv\n Construct settings are: \n {self.construct_settings} \n"
            + f" Config settings are:\n {self.config_settings} \n Initial value:\n {self.initial_value}\n"
        )

        logger.debug(debug_str)

        if (
            self.construct_settings["valtype"] not in ["s", "e"]
            and isinstance(self.initial_value, collections.abc.Sequence)
            and not self.construct_settings["valtype"].startswith("a")
        ):
            self.construct_settings["valtype"] = "a" + self.construct_settings["valtype"]

        if self.pvtype == PVTypes.ENUM:
            self.construct_settings.pop("valtype")
            nt = NTEnum(**self.construct_settings)
        else:
            nt = NTScalar(**self.construct_settings)

        self._config_timestamp()

        pvobj = SharedNT(nt=nt, initial={"value": self.initial_value, **self.config_settings})
        # handler._name = pv_name

        if self.read_only:
            pvobj.handler.read_only = True

        return pvobj

    def copy(self) -> BasePVRecipe:
        """Return a shallow copy of this instance"""
        return dataclasses.replace(self)

    def set_timestamp(self, timestamp: float):
        """Set the timestamp, floating point number in seconds since epoch"""
        self.timestamp = Timestamp(timestamp)


class PVScalarRecipe(BasePVRecipe, ABC):
    """Recipe to build an NTScalar"""

    def __post_init__(self):
        super().__post_init__()
        if self.pvtype != PVTypes.DOUBLE and self.pvtype != PVTypes.INTEGER and self.pvtype != PVTypes.STRING:
            raise ValueError(f"Unsupported pv type {self.pvtype} for class {{self.__class__.__name__}}")

    def set_control_limits(self, low: Numeric | None = None, high: Numeric | None = None, min_step=0):
        """
        Add control limits
        config is a dictionary of low_limit and high_limit. This is used by the config_reader.
        """
        if self.pvtype == PVTypes.DOUBLE:
            if low is None:
                low = MIN_FLOAT
            if high is None:
                high = MAX_FLOAT
            self.control = Control[float](limit_low=low, limit_high=high, min_step=min_step)
        elif self.pvtype == PVTypes.INTEGER:
            if low is None:
                low = MIN_INT32
            if high is None:
                high = MAX_INT32
            self.control = Control[int](limit_low=low, limit_high=high, min_step=min_step)
        elif self.pvtype == PVTypes.STRING:
            raise SyntaxError("Control limits not supported on string PVs")
        else:
            raise ValueError("Unknown pvtype")

    def set_display_limits(
        self,
        low: Numeric | None = None,
        high: Numeric | None = None,
        units: str = "",
        format: Format = Format.DEFAULT,
        precision: int = 2,
    ):
        """
        Add display limits
        config is a dictionary of low_limit and high_limit. This is used by the config_reader.
        """
        if isinstance(format, str):
            # check if it's in the available options
            choices = "UNINITIATED"
            try:
                choices = [form.value[1] for form in Format]
                idx = choices.index(format.title())
                format = list(Format)[idx]
            except ValueError as e:
                raise ValueError(f"{format} not an available format, choices are: {choices}") from e

        if self.pvtype == PVTypes.DOUBLE:
            if low is None:
                low = MIN_FLOAT
            if high is None:
                high = MAX_FLOAT
            self.display = Display[float](
                limit_low=low,
                limit_high=high,
                units=units,
                format=format,
                precision=precision,
            )
        elif self.pvtype == PVTypes.INTEGER:
            if low is None:
                low = MIN_INT32
            if high is None:
                high = MAX_INT32
            self.display = Display[int](
                limit_low=low,
                limit_high=high,
                units=units,
                format=format,
                precision=precision,
            )
        elif self.pvtype == PVTypes.STRING:
            raise SyntaxError("Display limits not supported on string PVs")
        else:
            raise ValueError("Unknown pvtype")

    def set_alarm_limits(
        self,
        low_warning: Numeric | None = None,
        high_warning: Numeric | None = None,
        low_alarm: Numeric | None = None,
        high_alarm: Numeric | None = None,
    ):
        """
        Add display limits
        config is a dictionary of low_limit and high_limit. This is used by the config_reader.
        """
        if self.pvtype == PVTypes.DOUBLE:
            if low_warning is None:
                low_warning = MIN_FLOAT
            if high_warning is None:
                high_warning = MAX_FLOAT
            if low_alarm is None:
                low_alarm = MIN_FLOAT
            if high_alarm is None:
                high_alarm = MAX_FLOAT
            self.alarm_limit = AlarmLimit[float](
                low_alarm_limit=low_alarm,
                low_warning_limit=low_warning,
                high_warning_limit=high_warning,
                high_alarm_limit=high_alarm,
            )
        elif self.pvtype == PVTypes.INTEGER:
            if low_warning is None:
                low_warning = MIN_INT32
            if high_warning is None:
                high_warning = MAX_INT32
            if low_alarm is None:
                low_alarm = MIN_INT32
            if high_alarm is None:
                high_alarm = MAX_INT32
            self.alarm_limit = AlarmLimit[float](
                low_alarm_limit=low_alarm,
                low_warning_limit=low_warning,
                high_warning_limit=high_warning,
                high_alarm_limit=high_alarm,
            )
        elif self.pvtype == PVTypes.STRING:
            raise SyntaxError("Alarm limits not supported on string PVs")
        else:
            raise ValueError("Unknown pvtype")

    def _config_display(self):
        # we configure the display settings if a Display object is configured or if all of
        # units, format and precision are not configured as the defaults
        if self.display:
            self.construct_settings["display"] = True
            self.construct_settings["form"] = True
            self.config_settings["display.description"] = self.description
            self.config_settings["display.units"] = self.display.units
            self.config_settings["display.precision"] = self.display.precision
            self.config_settings["display.form.index"] = self.display.format.value[0]
            self.config_settings["display.form.choices"] = [form.value[1] for form in Format]
            self.config_settings["display.limitLow"] = self.display.limit_low
            self.config_settings["display.limitHigh"] = self.display.limit_high

    def _config_alarm_limit(self):
        if self.alarm_limit:
            self.construct_settings["valueAlarm"] = True
            self.config_settings["valueAlarm.active"] = self.alarm_limit.active
            self.config_settings["valueAlarm.lowAlarmLimit"] = self.alarm_limit.low_alarm_limit
            self.config_settings["valueAlarm.lowWarningLimit"] = self.alarm_limit.low_warning_limit
            self.config_settings["valueAlarm.highWarningLimit"] = self.alarm_limit.high_warning_limit
            self.config_settings["valueAlarm.highAlarmLimit"] = self.alarm_limit.high_alarm_limit
            self.config_settings["valueAlarm.lowAlarmSeverity"] = self.alarm_limit.low_alarm_severity.value
            self.config_settings["valueAlarm.lowWarningSeverity"] = self.alarm_limit.low_warning_severity.value
            self.config_settings["valueAlarm.highWarningSeverity"] = self.alarm_limit.high_warning_severity.value
            self.config_settings["valueAlarm.highAlarmSeverity"] = self.alarm_limit.high_alarm_severity.value
            self.config_settings["valueAlarm.hysteresis"] = self.alarm_limit.hysteresis

    def _config_control(self):
        if self.control:
            self.construct_settings["control"] = True
            self.config_settings["control.limitLow"] = self.control.limit_low
            self.config_settings["control.limitHigh"] = self.control.limit_high
            self.config_settings["control.minStep"] = self.control.min_step



from p4pillon.utils import time_in_seconds_and_nanoseconds


def test_time_in_seconds_and_nanoseconds():
    seconds, nanoseconds = time_in_seconds_and_nanoseconds(123.456)
    assert seconds == 123
    assert nanoseconds == 456000000



"""
SimpleServer is used to create PVs and manage their lifetimes
"""

from __future__ import annotations

import logging

from p4p.client.thread import Context
from p4p.server import Server, StaticProvider

from p4pillon.pvrecipe import BasePVRecipe
from p4pillon.server.simpleserver import BaseSimpleServer
from p4pillon.server.thread import SharedPV

logger = logging.getLogger(__name__)


class SimpleServer(BaseSimpleServer):
    """
    Creates PVs and manages their lifetimes
    """

    def __init__(self, prefix: str = "") -> None:
        """
        Initialize the SimpleServer instance.

        :param prefix: The prefix to be added to the PVs (Process Variables) of the server e.g. DEV: Defaults to "".
        """
        super().__init__(prefix)

        self._provider = StaticProvider()
        self._server: Server | None = None

        self._running = False

        self._ctxt = Context("pva")

    def start(self) -> None:
        """Start the SimpleServer"""

        # iterate over all the PVs and initialise them if they haven't
        # been already, add them to the provider and start the server
        # this means that PVs are only 'opened' and given a time stamp
        # at the time the server itself is started
        for pv_name, pv in self._pvs.items():
            self._provider.add(pv_name, pv)

        self._server = Server(providers=[self._provider])

        logger.debug("Started Server with %s", self.pvlist)

        self._running = True

    def stop(self) -> None:
        """Stop the SimpleServer"""

        # iterate over all the PVs and close them before removing them
        # from the provider and closing the server
        for pv_name, pv in self._pvs.items():
            pv.close()
            self._provider.remove(pv_name)
        if self._server:
            self._server.stop()
        logger.debug("\nStopped server")

        self._running = False

    def add_pv(self, pv_name: str, pv_recipe: BasePVRecipe) -> SharedPV:
        """
        Add a PV to the server

        :param pv_name: The name of the PV to be added.
        :param pv_recipe: The recipe with instructions for creating the PV.
        :return: The created PV.
        """

        if not pv_name.startswith(self.prefix):
            pv_name = self.prefix + pv_name
        returnval = self._pvs[pv_name] = pv_recipe.create_pv(pv_name)

        # If the server is already running then we need to add this PV to
        # the live system
        if self._running:
            self._provider.add(pv_name, returnval)
            logger.debug("Added %s to server", pv_name)

        return returnval

    def remove_pv(self, pv_name: str) -> None:
        """
        Remove a PV from the server

        :param pv_name: The name of the PV to be removed.
        :raises KeyError: If the PV is not found in the list managed by the server.
        """

        if not pv_name.startswith(self.prefix):
            pv_name = self.prefix + pv_name

        # TODO: Consider the implications if this throws an exception
        pv = self._pvs.pop(pv_name)
        pv.close()
        if self._running:
            # If the server is already running then we need to remove this PV
            # from the live system
            self._provider.remove(pv_name)
        logger.debug("Removed %s from server", pv_name)



from p4p.server import Server

from p4pillon.nt import NTEnum
from p4pillon.thread.sharednt import SharedNT

pv = SharedNT(
    nt=NTEnum(display=True),  # scalar double
    initial={
        "value.index": 0,
        "value.choices": ["STOP", "START", "STANDBY"],
        "display.description": "Pump on/off control word.",
    },
)  # setting initial value also open()'s

Server.forever(
    providers=[
        {
            "demo:pv:name": pv,
        }
    ]
)  # runs until KeyboardInterrupt



"""
A simple example of using the SharedNT class to setup a set and readback from
"simulated" hardware.
"""

import random
from collections import OrderedDict
from time import sleep

from p4p import Value
from p4p.nt import NTScalar
from p4p.server import Server, ServerOperation

from p4pillon.nthandlers import Handler
from p4pillon.thread.sharednt import SharedNT


class SimulatedHardware:
    """
    Our simulated hardware is simple a gaussian random number generator
    where we change the mean value.
    """

    def __init__(self, value=0.0):
        self.value = value

    def poll(self):
        """
        Return a different random value each time the "hardware" is polled.
        """
        return self.value + random.gauss(0, 2.0)


class UserReportHandler(Handler):
    """
    Simply prints the user account attempting to perform a put operation.
    """

    def put(self, pv: SharedNT, op: ServerOperation):
        print(f"Operation attempted by user {op.account()} on pv {op.name()}")


class HWWriteHandler(Handler):
    """
    An example of a handler that would write to hardware.
    Note that it is important that this be the last handler in the chain, otherwise
    important functionality like control limits will not be applied.
    TODO: This should actually be the second last handler, the timestamp handler
          ought to be last, but the current design does not allow that.
    """

    def __init__(self, hardware: SimulatedHardware):
        self.hardware = hardware

    def post(self, pv: SharedNT, value: Value):
        if value.changed("value"):
            self.hardware.value = value["value"]

    def put(self, pv: SharedNT, op: ServerOperation):
        pass


def main():
    """
    Construct the Normative Type PVs, start the server, and update the hardware
    readback every 0.5 seconds.
    """

    hw = SimulatedHardware(4.5)  # Set initial value of simulated hardware

    # Create a SharedNT for readback from the hardware and another for the setpoint.
    # The readback is read-only, the setpoint is writable.
    # The readback is set to trigger an alarm if its value is over 17.
    # The setpoint is limited to a maximum value of 25.
    pv_hwget = SharedNT(
        nt=NTScalar(
            "d",
            valueAlarm=True,
        ),
        initial={
            "value": hw.poll(),
            "valueAlarm.active": True,
            "valueAlarm.highAlarmLimit": 17,
            "valueAlarm.highAlarmSeverity": 2,  # Not obvious, but without this the highAlarmLimit above will not work
        },
    )
    pv_hwget.handler.read_only = True  # Make the readback read-only

    pv_hwset = SharedNT(
        nt=NTScalar(
            "d",
            control=True,
        ),  # scalar double
        initial={"value": hw.poll(), "control.limitHigh": 25},
        auth_handlers=OrderedDict({"spy": UserReportHandler()}),
        user_handlers=OrderedDict({"hwwrite": HWWriteHandler(hw)}),
    )

    pvs = {
        "demo:hw_set": pv_hwset,  # PV name only appears here
        "demo:hw_get": pv_hwget,  # PV name only appears here
    }

    with Server(
        providers=[pvs],
    ):
        print("Server starting. Press Ctrl+C (or equivalent) to stop.")
        print("Server is providing PVs: ")
        for pv_name, pv in pvs.items():
            print(f"  {pv_name} with handlers {list(pv.handler.keys())}")

        try:
            while True:
                pv_hwget.post({"value": hw.poll()})
                sleep(0.5)

        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    main()



"""
Integration tests for expected behaviour of NTEnum PV types:
- [x] creation / modification (values, descriptions)
- [ ] alarm handling - match alarm
- [ ] alarm handling - invalid alarm
- [ ] calc records
- [ ] forward linking records
"""

import sys
from pathlib import Path

import pytest
import yaml
from helpers import put_different_value_enum, put_metadata

from tests.integration.thread.assertions import (
    assert_enum_value_changed,
    assert_enum_value_not_changed,
)

root_dir = Path(__file__).parents[2]

sys.path.append(str(root_dir))


with open(f"{root_dir}/integration/ntenum_config.yml") as f:
    ntenum_config = yaml.load(f, Loader=yaml.SafeLoader)
    f.close()


@pytest.mark.parametrize("pvname, pv_config", list(ntenum_config.items()))
def test_configs(pvname, enum_yaml_server, pv_config, ctx):
    # NOTE by using pytest and parameterize here we run the test individually
    # per PV in the config file, helping us to identify which PVs are causing
    # problems (this would be much more difficult if we were iterating over
    # a list from within the same test)
    pvname = enum_yaml_server.prefix + pvname

    assert pvname in enum_yaml_server.pvlist

    pv_state = ctx.get(pvname).raw.todict()

    assert pv_state.get("descriptor", "") == pv_config.get("description", "")

    assert pv_state.get("display") is None
    assert pv_state.get("control") is None
    assert pv_state.get("valueAlarm") is None


@pytest.mark.parametrize("pvname, pv_config", list(ntenum_config.items()))
def test_value_change(pvname, enum_yaml_server, pv_config, ctx):
    pvname = enum_yaml_server.prefix + pvname

    if not pv_config.get("read_only"):
        put_val, put_timestamp = put_different_value_enum(ctx, pvname)
        assert_enum_value_changed(pvname, put_val, put_timestamp, ctx)
    else:
        pytest.xfail(
            "Unsure on expected behaviour - expect the RemoteError or continue \
                     working with a warning logged to user"
        )
        assert_enum_value_not_changed(pvname, put_val, ctx)


@pytest.mark.parametrize("pvname, pv_config", list(ntenum_config.items()))
def test_field_change(pvname, enum_yaml_server, pv_config, ctx):
    pvname = enum_yaml_server.prefix + pvname

    current_description = ctx.get(pvname).raw.todict().get("descriptor")
    new_description = current_description + " modified"

    if not pv_config.get("read_only"):
        put_timestamp = put_metadata(ctx, pvname, "descriptor", new_description)

        pvstate = ctx.get(pvname)

        assert pvstate.raw.todict().get("descriptor") == new_description
        assert pvstate.timestamp >= put_timestamp

    else:
        pytest.xfail("Unsure on expected behaviour for read-only field changes")



"""
Monkey patch in required changes to NTEnum
"""

# pylint: disable=unused-wildcard-import, wildcard-import
from p4p import version as __p4p_version
from p4p.nt.enum import *  # pyright: ignore[reportWildcardImportFromLibrary] # noqa: F403
from p4p.nt.enum import NTEnum
from p4p.nt.scalar import ntwrappercommon
from p4p.wrapper import Value

if __p4p_version > "4.2.1":
    pass
else:

    def __wrap(self, value, choices=None, **kws):
        """Pack python value into Value

        Accepts dict to explicitly initialize fields by name.
        Any other type is assigned to the 'value' field via
        the self.assign() method.
        """
        if isinstance(value, Value):
            pass
        elif isinstance(value, ntwrappercommon):
            kws.setdefault("timestamp", value.timestamp)
            value = value.raw
        elif isinstance(value, dict):
            # if index, choices not in value.keys(), then
            # use value dict to initalize fields by name
            if {"index", "choices"}.isdisjoint(value):
                value = self.Value(self.type, value)
            # if value = {'index': ..., 'choices': ...}, then
            # assign these to value.index, value.choices
            else:
                value = self.Value(self.type, {"value": value})
        else:
            # index or string
            V = self.type()
            if choices is not None:
                V["value.choices"] = choices
            self.assign(V, value)
            value = V

        # pylint: disable=W0212
        self._choices = value["value.choices"] or self._choices  # pyright: ignore[reportOptionalSubscript]
        return self._annotate(value, **kws)  # pylint: disable=W0212

    NTEnum.Value = Value  # pyright: ignore[reportAttributeAccessIssue]
    NTEnum.wrap = __wrap



from unittest.mock import MagicMock

import numpy as np
import pytest

from p4pillon.definitions import PVTypes
from p4pillon.thread.config_reader import parse_config
from p4pillon.thread.pvrecipe import PVEnumRecipe, PVScalarArrayRecipe, PVScalarRecipe


@pytest.mark.parametrize(
    "name, config, pvtype",
    [
        ("TEST:PV:DOUBLE", {"initial": 0.0, "type": "DOUBLE"}, PVTypes.DOUBLE),
        ("TEST:PV:DOUBLE", {"initial": 17.0, "type": "DOUBLE"}, PVTypes.DOUBLE),
        ("TEST:PV:INTEGER", {"initial": 0, "type": "INTEGER"}, PVTypes.INTEGER),
        ("TEST:PV:DOUBLE", {"type": "DOUBLE"}, PVTypes.DOUBLE),
        ("TEST:PV:INTEGER", {"type": "INTEGER"}, PVTypes.INTEGER),
        ("TEST:PV:DOUBLE:ARRAY", {"type": "DOUBLE", "array_size": 3}, PVTypes.DOUBLE),
        ("TEST:PV:INTEGER:ARRAY", {"type": "INTEGER", "array_size": 3}, PVTypes.INTEGER),
    ],
)
def test_parse_config_with_server(name, config, pvtype):
    server = MagicMock()

    config["description"] = "test"

    recipes = parse_config({name: config}, server)
    assert len(recipes) == 1

    assert recipes[name].description == config["description"]
    assert recipes[name].pvtype == pvtype

    if config.get("array_size"):
        assert isinstance(recipes[name], PVScalarArrayRecipe)
        assert np.allclose(recipes[name].initial_value, np.array([0, 0, 0]))
    else:
        assert isinstance(recipes[name], PVScalarRecipe)
        assert recipes[name].initial_value == config.get("initial", 0.0)

    assert server.add_pv.call_args_list[0][0][0] == name


@pytest.mark.parametrize(
    "config, error_msg",
    [
        ({"description": "test"}, "'type' not specified"),
        ({"type": PVTypes.DOUBLE}, "'description' not specified"),
    ],
)
def test_parse_config_syntax_errors(config, error_msg):
    with pytest.raises(SyntaxError) as e:
        parse_config({"TEST:PV": config})
    assert error_msg in str(e)


@pytest.mark.parametrize(
    "name, config, pvtype",
    [
        ("TEST:PV:ENUM", {"initial": {"index": 0, "choices": ["False", "True"]}, "type": "ENUM"}, PVTypes.ENUM),
        ("TEST:PV:ENUM", {"initial": {"index": 1, "choices": ["OFF", "ON"]}, "type": "ENUM"}, PVTypes.ENUM),
    ],
)
def test_parse_config_with_server_enum(name, config, pvtype):
    server = MagicMock()

    config["description"] = "test"

    recipes = parse_config({name: config}, server)
    assert len(recipes) == 1

    assert isinstance(recipes[name], PVEnumRecipe)
    assert recipes[name].description == config["description"]
    assert recipes[name].pvtype == pvtype

    assert server.add_pv.call_args_list[0][0][1].initial_value == config["initial"]
    assert isinstance(server.add_pv.call_args_list[0][0][1], PVEnumRecipe)
    assert server.add_pv.call_args_list[0][0][0] == name



"""Enums used by p4pillon"""

from enum import Enum, IntEnum


class PVTypes(Enum):
    """Supported p4p Types"""

    DOUBLE = "d"
    INTEGER = "i"
    STRING = "s"
    ENUM = "e"


class AlarmSeverity(IntEnum):
    """PVAccess Normative Type alarm severities"""

    NO_ALARM = 0
    MINOR_ALARM = 1
    MAJOR_ALARM = 2
    INVALID_ALARM = 3
    UNDEFINED_ALARM = 4


class AlarmStatus(IntEnum):
    """PVAccess Normative Type alarm status codes"""

    NO_STATUS = 0
    DEVICE_STATUS = 1
    DRIVER_STATUS = 2
    RECORD_STATUS = 3
    DB_STATUS = 4
    CONF_STATUS = 5
    UNDEFINED_STATUS = 6
    CLIENT_STATUS = 7


class Format(Enum):
    """PVAccess Normative Type display format mappings"""

    DEFAULT = (0, "Default")
    STRING = (1, "String")
    BINARY = (2, "Binary")
    DECIMAL = (3, "Decimal")
    HEX = (4, "Hex")
    EXPONENTIAL = (5, "Exponential")
    ENGINEERING = (6, "Engineering")


MIN_FLOAT = float("-inf")
MAX_FLOAT = float("inf")
MIN_INT32 = -2147483648
MAX_INT32 = 2147483647



"""
Integration tests for expected behaviour of NTScalar PV types:
- [x] creation / modification (values, descriptions, limits)
- [x] alarm handling
- [x] control handling
- [ ] calc records
- [ ] forward linking records
"""

import time
from pathlib import Path

import pytest
import yaml
from helpers import put_different_value_scalar, put_metadata
from p4p._p4p import RemoteError
from p4p.client.thread import Context

from p4pillon.definitions import PVTypes
from p4pillon.thread.pvrecipe import PVScalarArrayRecipe, PVScalarRecipe
from p4pillon.thread.server import SimpleServer
from tests.integration.thread.assertions import (
    assert_correct_alarm_config,
    assert_correct_control_config,
    assert_correct_display_config,
    assert_pv_in_major_alarm_state,
    assert_pv_in_minor_alarm_state,
    assert_pv_not_in_alarm_state,
    assert_value_changed,
    assert_value_not_changed,
)

root_dir = Path(__file__).parents[2]


with open(f"{root_dir}/integration/ntscalar_config.yml") as f:
    ntscalar_config = yaml.load(f, Loader=yaml.SafeLoader)
    f.close()


@pytest.mark.parametrize("pvname, pv_config", list(ntscalar_config.items()))
def test_configs(pvname, yaml_server, pv_config, ctx):
    # NOTE by using pytest and parameterize here we run the test individually
    # per PV in the config file, helping us to identify which PVs are causing
    # problems (this would be much more difficult if we were iterating over
    # a list from within the same test)
    pvname = yaml_server.prefix + pvname

    pv_type = pv_config["type"]
    pv_is_numeric = pv_type in [PVTypes.DOUBLE.name, PVTypes.INTEGER.name]

    assert pvname in yaml_server.pvlist

    pv_state = ctx.get(pvname).raw.todict()
    # if we only provide a description with no other display fields, only
    # descriptor will be present but it should be in all PVs. Whereas when
    # any other field is specified like units etc the display.description
    # should also be configured
    assert pv_state.get("descriptor", "") == pv_config.get("description", "")

    if pv_is_numeric:
        if "display" in pv_config.keys():
            assert_correct_display_config(pv_state, pv_config)
        if "control" in pv_config.keys():
            assert_correct_control_config(pv_state, pv_config)
        if "valueAlarm" in pv_config.keys():
            assert_correct_alarm_config(pv_state, pv_config)

    else:
        assert pv_state.get("display") is None
        assert pv_state.get("control") is None
        assert pv_state.get("valueAlarm") is None


@pytest.mark.parametrize("pvname, pv_config", list(ntscalar_config.items()))
def test_value_change(pvname, yaml_server, pv_config, ctx):
    pvname = yaml_server.prefix + pvname

    current_state = ctx.get(pvname)

    if not pv_config.get("read_only"):
        put_val, put_timestamp = put_different_value_scalar(ctx, pvname)
        assert_value_changed(pvname, put_val, put_timestamp, ctx)
    else:
        with pytest.raises(RemoteError) as e:
            put_different_value_scalar(ctx, pvname)

        assert "read-only" in str(e)
        pvstate = ctx.get(pvname)

        assert pvstate.timestamp == current_state.timestamp
        assert pvstate == current_state


@pytest.mark.parametrize("pvname, pv_config", list(ntscalar_config.items()))
def test_field_change(pvname, yaml_server, pv_config, ctx):
    pvname = yaml_server.prefix + pvname

    current_state = ctx.get(pvname)

    current_description = current_state.raw.todict().get("descriptor")
    new_description = current_description + " modified"

    if not isinstance(pv_config.get("initial"), list):
        if not pv_config.get("read_only"):
            put_timestamp = put_metadata(ctx, pvname, "descriptor", new_description)
            pvstate = ctx.get(pvname)

            assert pvstate.raw.todict().get("descriptor") == new_description
            assert pvstate.timestamp >= put_timestamp
        else:
            with pytest.raises(RemoteError) as e:
                put_metadata(ctx, pvname, "descriptor", new_description)
            assert "read-only" in str(e)
            pvstate = ctx.get(pvname)
            assert pvstate.timestamp == current_state.timestamp
            assert pvstate.raw.todict().get("descriptor") == current_description
    else:
        pytest.xfail(reason="Currently unable to change fields in NTScalarArrays")


def test_alarm_limit_change_readonly(basic_server, ctx):
    # for each PV that is alarmed, we change the upper alarm limit on the PV
    # and check if setting the value to something above/below that triggers
    # the correct alarm state
    pvname = "TEST:ALARM:LIMIT:PV"

    alarm_config = {
        "low_alarm": -9,
        "low_warning": -4,
        "high_warning": 4,
        "high_alarm": 9,
    }
    pv_double1 = PVScalarRecipe(PVTypes.DOUBLE, "An example alarmed PV", 0)
    pv_double1.set_alarm_limits(**alarm_config)
    basic_server.add_pv(pvname, pv_double1)

    # TODO: This is a very messy way of making a rule not read_only!
    basic_server[pvname]._handler["alarm_limit"].read_only = True

    basic_server.start()

    ctx.put(pvname, -5)
    assert_pv_in_minor_alarm_state(pvname, ctx)

    put_metadata(ctx, pvname, "valueAlarm.lowWarningLimit", -6)
    assert_pv_in_minor_alarm_state(pvname, ctx)

    ctx.put(pvname, -10)
    assert_pv_in_major_alarm_state(pvname, ctx)

    put_metadata(ctx, pvname, "valueAlarm.lowAlarmLimit", -11)
    assert_pv_in_major_alarm_state(pvname, ctx)


def test_alarm_limit_change(basic_server, ctx):
    # for each PV that is alarmed, we change the upper alarm limit on the PV
    # and check if setting the value to something above/below that triggers
    # the correct alarm state
    pvname = "TEST:ALARM:LIMIT:PV"

    alarm_config = {
        "low_alarm": -9,
        "low_warning": -4,
        "high_warning": 4,
        "high_alarm": 9,
    }
    pv_double1 = PVScalarRecipe(PVTypes.DOUBLE, "An example alarmed PV", 0)
    pv_double1.set_alarm_limits(**alarm_config)
    basic_server.add_pv(pvname, pv_double1)

    # TODO: This is a very messy way of making a rule not read_only!
    basic_server[pvname]._handler["alarm_limit"].read_only = False

    basic_server.start()

    ctx.put(pvname, -5)
    assert_pv_in_minor_alarm_state(pvname, ctx)

    put_metadata(ctx, pvname, "valueAlarm.lowWarningLimit", -6)
    assert_pv_not_in_alarm_state(pvname, ctx)

    ctx.put(pvname, -10)
    assert_pv_in_major_alarm_state(pvname, ctx)

    put_metadata(ctx, pvname, "valueAlarm.lowAlarmLimit", -11)
    assert_pv_in_minor_alarm_state(pvname, ctx)


class TestAlarms:
    """Integration test case for validating alarm limit behaviour on a variety
    of PV types"""

    @pytest.mark.parametrize("pvtype", [(PVTypes.DOUBLE), (PVTypes.INTEGER)])
    def test_basic_alarm_logic(self, basic_server: SimpleServer, ctx: Context, pvtype):
        # here we have an example of a pretty standard range alarm configuration
        pvname = "TEST:ALARM:PV"

        alarm_config = {
            "low_alarm": -9,
            "low_warning": -4,
            "high_warning": 4,
            "high_alarm": 9,
        }
        pv_double1 = PVScalarRecipe(pvtype, "An example alarmed PV", 0)
        pv_double1.set_alarm_limits(**alarm_config)
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        ctx.put(pvname, -10)
        assert_pv_in_major_alarm_state(pvname, ctx)
        ctx.put(pvname, -5)
        assert_pv_in_minor_alarm_state(pvname, ctx)
        ctx.put(pvname, 0)
        assert_pv_not_in_alarm_state(pvname, ctx)
        ctx.put(pvname, 5)
        assert_pv_in_minor_alarm_state(pvname, ctx)
        ctx.put(pvname, 10)
        assert_pv_in_major_alarm_state(pvname, ctx)

    @pytest.mark.parametrize("pvtype", [(PVTypes.DOUBLE), (PVTypes.INTEGER)])
    def test_defaults_alarm_logic(self, basic_server, ctx: Context, pvtype):
        # PVs that use the default values will never go into the alarm state
        pvname = "TEST:ALARM:PV"

        pv_double1 = PVScalarRecipe(pvtype, "An example numeric alarmed PV", 0)
        pv_double1.set_alarm_limits()
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        for val in [-10, -5, 0, 5, 10]:
            ctx.put(pvname, val)
            assert_pv_not_in_alarm_state(pvname, ctx)

    @pytest.mark.parametrize("pvtype", [(PVTypes.DOUBLE), (PVTypes.INTEGER)])
    def test_only_high_alarm(self, basic_server, ctx: Context, pvtype: PVTypes):
        # PVs that use the default values will never go into the alarm state
        pvname = "TEST:ALARM:PV"

        pv_double1 = PVScalarRecipe(pvtype, "An example numeric alarmed PV", 0)
        pv_double1.set_alarm_limits(high_alarm=9)
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        for val in [-10, -5, 0, 5, 9, 10]:
            ctx.put(pvname, val)
            if val < 9:
                assert_pv_not_in_alarm_state(pvname, ctx)
            else:
                assert_pv_in_major_alarm_state(pvname, ctx)

    @pytest.mark.parametrize("pvtype", [(PVTypes.DOUBLE), (PVTypes.INTEGER)])
    def test_basic_alarm_logic_array_vals(self, basic_server: SimpleServer, ctx: Context, pvtype):
        # here we have an example of a pretty standard range alarm configuration but on
        # an array PV. In this case we expect the alarm to be triggered if ANY of the
        # values in the list exceed these values

        pvname = "TEST:ALARM:PV"

        alarm_config = {
            "low_alarm": -9,
            "low_warning": -4,
            "high_warning": 4,
            "high_alarm": 9,
        }
        pv_double1 = PVScalarArrayRecipe(pvtype, "An example alarmed PV", [0])
        pv_double1.set_alarm_limits(**alarm_config)
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        test_list = [0] * 5

        # At start
        ctx.put(pvname, [-10] + test_list)
        assert_pv_in_major_alarm_state(pvname, ctx)
        ctx.put(pvname, [-5] + test_list)
        assert_pv_in_minor_alarm_state(pvname, ctx)
        ctx.put(pvname, [0] + test_list)
        assert_pv_not_in_alarm_state(pvname, ctx)
        ctx.put(pvname, [5] + test_list)
        assert_pv_in_minor_alarm_state(pvname, ctx)
        ctx.put(pvname, [10] + test_list)
        assert_pv_in_major_alarm_state(pvname, ctx)

        # At end
        ctx.put(pvname, test_list + [-10])
        assert_pv_in_major_alarm_state(pvname, ctx)
        ctx.put(pvname, test_list + [-5])
        assert_pv_in_minor_alarm_state(pvname, ctx)
        ctx.put(pvname, test_list + [0])
        assert_pv_not_in_alarm_state(pvname, ctx)
        ctx.put(pvname, test_list + [5])
        assert_pv_in_minor_alarm_state(pvname, ctx)
        ctx.put(pvname, test_list + [10])
        assert_pv_in_major_alarm_state(pvname, ctx)

    @pytest.mark.parametrize("pvtype", [(PVTypes.DOUBLE), (PVTypes.INTEGER)])
    def test_defaults_alarm_logic_arrays(self, basic_server, ctx: Context, pvtype):
        # PVs that use the default values will never go into the alarm state
        pvname = "TEST:ALARM:PV"

        pv_double1 = PVScalarArrayRecipe(pvtype, "An example numeric alarmed PV", 0)
        pv_double1.set_alarm_limits()
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        test_list = [0] * 5

        for val in [-10, -5, 0, 5, 10]:
            ctx.put(pvname, test_list + [val])
            assert_pv_not_in_alarm_state(pvname, ctx)


class TestControl:
    """Integration test case for validating control limit behaviour on a variety
    of PV types"""

    @pytest.mark.parametrize(
        "pvtype, put_val, expected_val",
        [
            (PVTypes.DOUBLE, -10, -9),
            (PVTypes.DOUBLE, 0, 0),
            (PVTypes.DOUBLE, 10, 9),
            (PVTypes.INTEGER, -10, -9),
            (PVTypes.INTEGER, 0, 0),
            (PVTypes.INTEGER, 10, 9),
        ],
    )
    def test_basic_control_logic(self, basic_server: SimpleServer, ctx: Context, pvtype, put_val, expected_val):
        # here we have an example of a PV with control limits
        pvname = "TEST:CONTROL:PV"

        control_config = {"low": -9, "high": 9, "min_step": 1}
        pv_double1 = PVScalarRecipe(pvtype, "An example PV with control limits", 0)
        pv_double1.set_control_limits(**control_config)
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        timestamp = time.time()
        ctx.put(pvname, put_val)
        assert_value_changed(pvname, expected_val, timestamp, ctx)

    @pytest.mark.parametrize(
        "pvtype, put_val",
        [
            (PVTypes.DOUBLE, -10),
            (PVTypes.DOUBLE, 0),
            (PVTypes.DOUBLE, 10),
            (PVTypes.INTEGER, -10),
            (PVTypes.INTEGER, 0),
            (PVTypes.INTEGER, 10),
        ],
    )
    def test_default_control_logic(self, basic_server: SimpleServer, ctx: Context, pvtype, put_val):
        # here we have an example of a PV with default control limits applied
        pvname = "TEST:CONTROL:PV"

        pv_double1 = PVScalarRecipe(pvtype, "An example PV with default control limits", 0)
        pv_double1.set_control_limits()
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        timestamp = time.time()
        ctx.put(pvname, put_val)
        assert_value_changed(pvname, put_val, timestamp, ctx)

    @pytest.mark.parametrize(
        "pvtype",
        [
            (PVTypes.DOUBLE),
            (PVTypes.INTEGER),
        ],
    )
    def test_control_logic_min_step(self, basic_server: SimpleServer, ctx: Context, pvtype):
        # putting a new value less than the minimum step should prevent
        # the value being set
        pvname = "TEST:CONTROL:PV"

        control_config = {"low": -9, "high": 9, "min_step": 2}
        pv_double1 = PVScalarRecipe(pvtype, "An example PV with control limits", 0)
        pv_double1.set_control_limits(**control_config)

        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()
        assert ctx.get(pvname).real == 0
        # setting the value to 1 shouldn't work because it's less than the minimum step
        timestamp = time.time()
        new_val = 1
        ctx.put(pvname, new_val)
        assert_value_not_changed(pvname, new_val, ctx)
        # but putting a value of 3 should work because it's above the minimum step
        timestamp = time.time()
        new_val = 3
        ctx.put(pvname, new_val)
        assert_value_changed(pvname, new_val, timestamp, ctx)

    @pytest.mark.parametrize(
        "pvtype, put_val, expected_val",
        [
            (PVTypes.DOUBLE, -10, -9),
            (PVTypes.DOUBLE, 0, 0),
            (PVTypes.DOUBLE, 10, 9),
            (PVTypes.INTEGER, -10, -9),
            (PVTypes.INTEGER, 0, 0),
            (PVTypes.INTEGER, 10, 9),
        ],
    )
    def test_basic_control_logic_array(self, basic_server: SimpleServer, ctx: Context, pvtype, put_val, expected_val):
        # here we have an example of a PV with control limits
        pvname = "TEST:CONTROL:PV"

        array_length = 6

        control_config = {"low": -9, "high": 9, "min_step": 1}
        pv_double1 = PVScalarArrayRecipe(pvtype, "An example array PV with control limits", [0] * array_length)
        pv_double1.set_control_limits(**control_config)
        basic_server.add_pv(pvname, pv_double1)

        basic_server.start()

        test_list = [0] * (array_length - 1)

        timestamp = time.time()
        ctx.put(pvname, test_list + [put_val])
        assert_value_changed(pvname, test_list + [expected_val], timestamp, ctx)



"""
Wrapper to SharedPV in p4p to automatically create
"""

from __future__ import annotations

import logging
from collections import OrderedDict

from p4pillon.composite_handler import CompositeHandler
from p4pillon.nt import NTBase, NTEnum, NTScalar
from p4pillon.nthandlers import ComposeableRulesHandler
from p4pillon.rules import (
    AlarmRule,
    ControlRule,
    ScalarToArrayWrapperRule,
    TimestampRule,
    ValueAlarmRule,
)
from p4pillon.server.raw import Handler
from p4pillon.server.thread import SharedPV

logger = logging.getLogger(__name__)


class SharedNT(SharedPV):
    """
    SharedNT is a wrapper around SharedPV that automatically adds handler
    functionality to support Normative Type logic.
    """

    def __init__(
        self,
        auth_handlers: OrderedDict[str, Handler] | None = None,
        user_handlers: OrderedDict[str, Handler] | None = None,
        **kws,
    ):
        # Check if there is a handler specified in the kws, and if not override it
        # with an NT handler.

        # Create a CompositeHandler. If there is no user supplied handler, and this is not
        # an NT type then it won't do anything. But it will still represent a stable interface

        if auth_handlers:
            handler = CompositeHandler(auth_handlers)
        else:
            handler = CompositeHandler()

        if "nt" in kws:
            nt: NTBase = kws["nt"]

            match nt:
                case NTScalar():
                    nttype_str: str = nt.type.getID()
                    if nttype_str.startswith("epics:nt/NTScalarArray"):
                        handler["control"] = ComposeableRulesHandler(ScalarToArrayWrapperRule(ControlRule()))
                        handler["alarm"] = ComposeableRulesHandler(
                            AlarmRule()
                        )  # ScalarToArrayWrapperRule unnecessary - no access to values
                        handler["alarm_limit"] = ComposeableRulesHandler(ScalarToArrayWrapperRule(ValueAlarmRule()))
                        handler["timestamp"] = ComposeableRulesHandler(TimestampRule())
                    elif nttype_str.startswith("epics:nt/NTScalar"):
                        handler["control"] = ComposeableRulesHandler(ControlRule())
                        handler["alarm"] = ComposeableRulesHandler(AlarmRule())
                        handler["alarm_limit"] = ComposeableRulesHandler(ValueAlarmRule())
                        handler["timestamp"] = ComposeableRulesHandler(TimestampRule())
                    else:
                        raise TypeError(f"Unrecognised NT type: {nttype_str}")
                case NTEnum():
                    handler["timestamp"] = ComposeableRulesHandler(TimestampRule())
                case _:
                    raise NotImplementedError(f"SharedNT does not support type: {nt.__class__.__name__}")

        if user_handlers:
            handler = handler | user_handlers
            handler.move_to_end("timestamp", last=True)  # Ensure timestamp is last

        kws["handler"] = handler

        super().__init__(**kws)

    @property
    def handler(self) -> CompositeHandler:
        return self._handler

    @handler.setter
    def handler(self, value: CompositeHandler):
        self._handler = value

    ## Disable handler decorators until we have a solid design.
    # Re-enable when / if possible

    @property
    def onFirstConnect(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    @property
    def onLastDisconnect(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    @property
    def on_open(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    @property
    def on_post(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    @property
    def put(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    @property
    def rpc(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    @property
    def on_close(self):
        raise NotImplementedError("Handler decorators are not currently compatible with multiple handlers.")

    ## Alternative PEP 8 comaptible handler decorators
    # @property
    # def on_first_connect(self):
    #     """Turn a function into an ISISHandler onFirstConnect() method."""

    #     def decorate(fn):
    #         self._handler.onFirstConnect = fn
    #         return fn

    #     return decorate

    # @property
    # def on_last_disconnect(self):
    #     """Turn a function into an ISISHandler onLastDisconnect() method."""

    #     def decorate(fn):
    #         self._handler.onLastDisconnect = fn
    #         return fn

    #     return decorate

    # @property
    # def on_put(self):
    #     """Turn a function into an ISISHandler put() method."""

    #     def decorate(fn):
    #         self._handler.put = fn
    #         return fn

    #     return decorate

    # @property
    # def on_rpc(self):
    #     """Turn a function into an ISISHandler rpc() method."""

    #     def decorate(fn):
    #         self._handler.rpc = fn
    #         return fn

    #     return decorate

    # @property
    # def on_post(self):
    #     """Turn a function into an ISISHandler post() method."""

    #     def decorate(fn):
    #         self._handler.post = fn
    #         return fn

    #     return decorate



"""
Composite Handler allows multiple standard handlers to be combined into a single handler.

And ordered dictionary is used to make the component handlers accessible by name.
The ordered dictionary also controls the order in which the handlers are called.
"""

from __future__ import annotations

from collections import OrderedDict

from p4p import Value
from p4p.server import ServerOperation

from p4pillon.server.raw import Handler, SharedPV


class HandlerException(Exception):
    """Exception raised for errors in the handler operations."""


class AbortHandlerException(HandlerException):
    """Exception raised to abort the current operation in the handler."""

    def __init__(self, message: str = "Operation aborted"):
        super().__init__(message)
        self.message = message


class CompositeHandler(Handler, OrderedDict):
    """Composite Handler for combining multiple component handlers into a single handler."""

    def __init__(self, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        Handler.__init__(self)

        self.read_only = False

    def open(self, value: Value):
        """Open all handlers in the composite handler."""
        for _name, handler in self.items():
            handler.open(value)

    def put(self, pv: SharedPV, op: ServerOperation):
        if self.read_only:
            errmsg = "This PV is read-only"
            op.done(error=errmsg)
            return

        errmsg = None

        for _name, handler in self.items():
            try:
                handler.put(pv, op)
            except AbortHandlerException as e:
                errmsg = e.message
                break

        if errmsg is None:
            pv.post(op.value())
            op.done()
        else:
            op.done(error=errmsg)

    def post(self, pv: SharedPV, value: Value):
        for _name, handler in self.items():
            handler.post(pv, value)

    def rpc(self, pv: SharedPV, op: ServerOperation):
        errmsg = None

        for handler in self.values():
            try:
                handler.rpc(pv, op)
            except AbortHandlerException as e:
                errmsg = e.message
                break

        op.done(error=errmsg)

    def on_first_connect(self, pv: SharedPV):
        """Called when the first client connects to the PV."""
        for handler in self.values():
            handler.onFirstConnect(pv)

    def onFirstConnect(self, pv: Value):
        self.on_first_connect(pv)

    def on_last_connect(self, pv: SharedPV):
        """Called when the last client channel is closed."""
        for handler in self.values():
            handler.onFirstConnect(pv)

    def onLastDisconnect(self, pv: Value):
        self.on_last_connect(pv)

    def close(self, pv: SharedPV):
        for handler in self.values():
            handler.close(pv)



import math
from unittest.mock import patch

import pytest
from p4p.nt import NTScalar

from p4pillon.composite_handler import CompositeHandler
from p4pillon.definitions import MAX_FLOAT, MAX_INT32, MIN_FLOAT, MIN_INT32, AlarmSeverity, Format, PVTypes
from p4pillon.thread.pvrecipe import PVEnumRecipe, PVScalarArrayRecipe, PVScalarRecipe


@pytest.mark.parametrize(
    "pvtype, display_config, expected_values",
    [
        (
            # passing an empty display dictionary gives the defaults
            PVTypes.INTEGER,
            {},
            (MIN_INT32, MAX_INT32, "", Format.DEFAULT, 2),
        ),
        (
            PVTypes.INTEGER,
            {"units": "V"},
            (MIN_INT32, MAX_INT32, "V", Format.DEFAULT, 2),
        ),
        (
            PVTypes.INTEGER,
            {"format": Format.ENGINEERING},
            (MIN_INT32, MAX_INT32, "", Format.ENGINEERING, 2),
        ),
        (
            PVTypes.DOUBLE,
            {"units": "V"},
            (MIN_FLOAT, MAX_FLOAT, "V", Format.DEFAULT, 2),
        ),
        (
            PVTypes.DOUBLE,
            {"low": -1.0, "high": 1.0},
            (-1.0, 1.0, "", Format.DEFAULT, 2),
        ),
        (
            PVTypes.DOUBLE,
            {"precision": 4},
            (MIN_FLOAT, MAX_FLOAT, "", Format.DEFAULT, 4),
        ),
        (
            PVTypes.DOUBLE,
            {"format": "ENGINEERING"},
            (MIN_FLOAT, MAX_FLOAT, "", Format.ENGINEERING, 2),
        ),
    ],
)
def test_ntscalar_display(pvtype, display_config, expected_values):
    for recipetype in [PVScalarRecipe, PVScalarArrayRecipe]:
        recipe = recipetype(pvtype, description="test PV", initial_value=0)

        assert recipe.display is None

        recipe.set_display_limits(**display_config)

        assert recipe.display.limit_low == expected_values[0]
        assert recipe.display.limit_high == expected_values[1]
        assert recipe.display.units == expected_values[2]
        assert recipe.display.format is expected_values[3]
        assert recipe.display.precision == expected_values[4]


@pytest.mark.parametrize(
    "pvtype, time_val",
    [
        pytest.param(
            PVTypes.INTEGER,
            123.456,
            marks=pytest.mark.xfail,
        ),  # Issue with _wrap preventing timestamp setting
        pytest.param(
            PVTypes.DOUBLE,
            123.456,
            marks=pytest.mark.xfail,
        ),  # Issue with _wrap preventing timestamp setting
        (PVTypes.INTEGER, None),
        (PVTypes.DOUBLE, None),
    ],
)
@patch("time.time", return_value=456.789)
def test_ntscalar_timestamp(mock_time, pvtype, time_val):
    for recipetype in [PVScalarRecipe, PVScalarArrayRecipe]:
        recipe = recipetype(pvtype, description="test PV", initial_value=0)

        assert recipe.timestamp is None

        if time_val is not None:
            recipe.set_timestamp(time_val)
            assert recipe.timestamp.time == time_val
        else:
            assert recipe.timestamp is None

        pv = recipe.create_pv("TEST:NAME")

        if time_val is not None:
            # once we've added the PVs and started the server, the PV timestamp should be respected
            print(pv.current().timestamp, time_val)
            assert math.isclose(pv.current().timestamp, time_val)
        else:
            # if the timestamp isn't set, we use the default time.time return val
            assert math.isclose(pv.current().timestamp, mock_time.return_value)


@pytest.mark.parametrize(
    "pvtype, control_config, expected_values",
    [
        (
            PVTypes.INTEGER,
            {},
            (
                MIN_INT32,
                MAX_INT32,
                0,
            ),
        ),
        (
            PVTypes.INTEGER,
            {"low": -5, "high": 5, "min_step": 1},
            (
                -5,
                5,
                1,
            ),
        ),
        (
            PVTypes.DOUBLE,
            {},
            (
                MIN_FLOAT,
                MAX_FLOAT,
                0,
            ),
        ),
        (
            PVTypes.DOUBLE,
            {"low": -5, "high": 5, "min_step": 0.1},
            (
                -5,
                5,
                0.1,
            ),
        ),
    ],
)
def test_ntscalar_control(pvtype, control_config, expected_values):
    for recipetype in [PVScalarRecipe, PVScalarArrayRecipe]:
        recipe = recipetype(pvtype, description="test PV", initial_value=0)

        assert recipe.control is None

        recipe.set_control_limits(**control_config)

        assert recipe.control.limit_low == expected_values[0]
        assert recipe.control.limit_high == expected_values[1]
        assert recipe.control.min_step == expected_values[2]


@pytest.mark.parametrize(
    "pvtype, alarm_config, expected_values",
    [
        (
            PVTypes.INTEGER,
            {},
            (MIN_INT32, MIN_INT32, MAX_INT32, MAX_INT32),
        ),
        (
            PVTypes.INTEGER,
            {"low_alarm": -5, "low_warning": -3, "high_alarm": 5, "high_warning": 3},
            (-5, -3, 3, 5),
        ),
        (
            # TODO confirm if this is the right behaviour?
            PVTypes.INTEGER,
            {
                "low_alarm": -5,
                "high_alarm": 5,
            },
            (-5, MIN_INT32, MAX_INT32, 5),
        ),
        (
            PVTypes.INTEGER,
            {"low_alarm": -5, "low_warning": 5},
            (-5, 5, MAX_INT32, MAX_INT32),
        ),
        (
            PVTypes.DOUBLE,
            {},
            (MIN_FLOAT, MIN_FLOAT, MAX_FLOAT, MAX_FLOAT),
        ),
        (
            PVTypes.DOUBLE,
            {"low_alarm": -5, "low_warning": -3, "high_alarm": 5, "high_warning": 3},
            (-5, -3, 3, 5),
        ),
        (
            # TODO confirm if this is the right behaviour?
            PVTypes.DOUBLE,
            {
                "low_alarm": -5,
                "high_alarm": 5,
            },
            (-5, MIN_FLOAT, MAX_FLOAT, 5),
        ),
        (
            PVTypes.DOUBLE,
            {"low_alarm": -5, "low_warning": 5},
            (-5, 5, MAX_FLOAT, MAX_FLOAT),
        ),
    ],
)
def test_ntscalar_alarm_limit(pvtype, alarm_config, expected_values):
    for recipetype in [PVScalarRecipe, PVScalarArrayRecipe]:
        recipe = recipetype(pvtype, description="test PV", initial_value=0)

        assert recipe.alarm_limit is None

        recipe.set_alarm_limits(**alarm_config)

        assert recipe.alarm_limit.low_alarm_limit == expected_values[0]
        assert recipe.alarm_limit.low_warning_limit == expected_values[1]
        assert recipe.alarm_limit.high_warning_limit == expected_values[2]
        assert recipe.alarm_limit.high_alarm_limit == expected_values[3]
        assert recipe.alarm_limit.low_alarm_severity == AlarmSeverity.MAJOR_ALARM
        assert recipe.alarm_limit.low_warning_severity == AlarmSeverity.MINOR_ALARM
        assert recipe.alarm_limit.high_warning_severity == AlarmSeverity.MINOR_ALARM
        assert recipe.alarm_limit.high_alarm_severity == AlarmSeverity.MAJOR_ALARM
        assert recipe.alarm_limit.hysteresis == 0


def test_ntscalar_string_errors():
    # string NTScalars don't support any of the standard numeric NTScalar fields like display,
    # control or alarm limits
    for recipetype in [PVScalarRecipe, PVScalarArrayRecipe]:
        recipe = recipetype(PVTypes.STRING, description="test PV", initial_value=0)

        # check display
        assert recipe.display is None
        with pytest.raises(SyntaxError) as e:
            recipe.set_display_limits()
        assert "not supported" in str(e)

        # check control
        assert recipe.control is None
        with pytest.raises(SyntaxError) as e:
            recipe.set_control_limits()
        assert "not supported" in str(e)

        # check valueAlarm
        assert recipe.alarm_limit is None
        with pytest.raises(SyntaxError) as e:
            recipe.set_alarm_limits()
        assert "not supported" in str(e)


def test_ntscalar_enum_error():
    with pytest.raises(ValueError) as e:
        PVScalarRecipe(PVTypes.ENUM, description="test", initial_value=1)

    assert "Unsupported pv type" in str(e)


@pytest.mark.parametrize(
    "recipe, pvtype, with_limits, expected_value",
    [
        (PVScalarRecipe, PVTypes.DOUBLE, True, 1),
        (PVScalarRecipe, PVTypes.INTEGER, True, 1),
        (PVScalarRecipe, PVTypes.DOUBLE, False, 1),
        (PVScalarRecipe, PVTypes.INTEGER, False, 1),
        (PVScalarArrayRecipe, PVTypes.DOUBLE, True, [1]),
        (PVScalarArrayRecipe, PVTypes.INTEGER, True, [1]),
        (PVScalarArrayRecipe, PVTypes.DOUBLE, False, [1]),
        (PVScalarArrayRecipe, PVTypes.INTEGER, False, [1]),
    ],
)
@patch("time.time")
def test_ntscalar_numeric_create_pv(mock_time, recipe, pvtype, with_limits, expected_value):
    mock_time.return_value = 123.456
    initial = 1.0
    recipe = recipe(pvtype, description="test", initial_value=initial)

    if with_limits:
        recipe.set_display_limits()
        recipe.set_control_limits()
        recipe.set_alarm_limits()

    pv = recipe.create_pv(pv_name="UNIT:TEST:PV")

    pvdict = pv.current().raw.todict()

    assert isinstance(pv._handler, CompositeHandler)
    assert list(pv._handler.keys()) == ["control", "alarm", "alarm_limit", "timestamp"]
    assert isinstance(pv.nt, NTScalar)  # change to check the name instead?
    assert pv.isOpen() is True

    assert pv.current().real == expected_value
    assert pv.current().timestamp == mock_time.return_value
    assert pvdict.get("alarm") is not None
    if with_limits:
        assert pvdict.get("display") is not None
        assert pvdict.get("control") is not None
        assert pvdict.get("valueAlarm") is not None
    else:
        assert pvdict.get("display") is None
        assert pvdict.get("control") is None
        assert pvdict.get("valueAlarm") is None


@pytest.mark.parametrize(
    "recipe, expected_value",
    [
        (PVScalarRecipe, "test"),
        # TODO work out how to fix this - currently failing with error:
        # ValueError: Unable to wrap ['test'] with <bound method NTScalar.wrap of <p4p.nt.scalar.NTScalar
        pytest.param(
            PVScalarArrayRecipe,
            ["test"],
            marks=pytest.mark.xfail,
        ),
    ],
)
@patch("time.time")
def test_ntscalar_string_create_pv(mock_time, recipe, expected_value):
    mock_time.return_value = 123.456
    initial = "test"
    recipe = recipe(PVTypes.STRING, description="test", initial_value=initial)

    pv = recipe.create_pv(pv_name="UNIT:TEST:PV")

    pvdict = pv.current().raw.todict()

    assert isinstance(pv._handler, CompositeHandler)
    assert list(pv._handler.keys()) == ["control", "alarm", "alarm_limit", "timestamp"]
    assert isinstance(pv.nt, NTScalar)
    assert pv.isOpen() is True
    assert pv.current().timestamp == mock_time.return_value
    assert pvdict["value"] == expected_value
    assert pvdict.get("alarm") is not None
    # string PVs shouldn't have any of these fields
    assert pvdict.get("display") is None
    assert pvdict.get("control") is None
    assert pvdict.get("valueAlarm") is None


@pytest.mark.parametrize(
    "pvtype",
    [(PVTypes.DOUBLE), (PVTypes.INTEGER), (PVTypes.STRING)],
)
def test_ntenum_bad_types(pvtype):
    with pytest.raises(ValueError) as e:
        PVEnumRecipe(pvtype, description="test enum", initial_value={"index": 0, "choices": ["OFF", "ON"]})

    assert "Unsupported pv type" in str(e)


@patch("time.time")
def test_ntenum_create_pv(mock_time):
    mock_time.return_value = 123.456

    recipe = PVEnumRecipe(PVTypes.ENUM, description="test enum", initial_value={"index": 0, "choices": ["OFF", "ON"]})

    pv = recipe.create_pv("TEST:PV:ENUM")

    pvdict = pv.current().raw.todict()

    assert pv.isOpen()
    assert isinstance(pv._handler, CompositeHandler)
    assert list(pv._handler.keys()) == ["timestamp"]
    assert pv.nt.type.getID() == "epics:nt/NTEnum:1.0"
    assert pv.isOpen() is True
    assert pv.current().timestamp == mock_time.return_value
    assert pvdict["value"] == {"index": 0, "choices": ["OFF", "ON"]}
    assert pvdict.get("alarm") is not None
    # enum PVs shouldn't have any of these fields
    assert pvdict.get("display") is None
    assert pvdict.get("control") is None
    assert pvdict.get("valueAlarm") is None


def test_ntenum_extras():
    recipe = PVEnumRecipe(PVTypes.ENUM, description="test enum", initial_value={"index": 0, "choices": ["OFF", "ON"]})

    # check display
    assert recipe.display is None
    with pytest.raises(AttributeError):
        recipe.set_display_limits()

    # check control
    assert recipe.control is None
    with pytest.raises(AttributeError):
        recipe.set_control_limits()

    # check valueAlarm
    assert recipe.alarm_limit is None
    with pytest.raises(AttributeError):
        recipe.set_alarm_limits()



"""
Monkey patch in required changes to Handlers and SharedPVs
"""

####
# First override the base class of p4pillon.server.thread.SharedPV with
# p4pillon.server.raw.SharedPV. This requires us to perform the imports
# in a very specific order, which means overriding Linter checks
import p4p.server.raw

from p4pillon.server.raw import SharedPV as _SharedPV

p4p.server.raw.SharedPV = _SharedPV

# pylint: disable=unused-import, wrong-import-order, wrong-import-position
from p4p.server.thread import Handler, SharedPV  # noqa: E402, F401,

#####
# Monkey patching the Handler is a simpler operation as it's a straight
# substitution with our new version.
# pylint: disable=ungrouped-imports
from p4pillon.server.raw import Handler as _Handler  # noqa: E402

Handler = _Handler  # noqa: F811
